"""Unit tests for all the methods for data management."""
import copy
import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import yaml

import qililab as ql
from qililab.data_management import load_results, save_platform, save_results
from qililab.platform import Platform
from tests.data import Galadriel
from tests.test_utils import build_platform


@patch("qililab.data_management.yaml.safe_load", return_value=copy.deepcopy(Galadriel.runcard))
@patch("qililab.data_management.open")
class TestPlatformData:
    """Unit tests for the `build_platform` function.."""

    def test_build_platform_passing_a_path_to_old_path_argument(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test build method."""
        with pytest.warns() as record:
            platform = ql.build_platform(path="_")
        assert isinstance(platform, Platform)
        assert len(record) == 1
        assert (
            str(record[0].message)
            == "`path` argument is deprecated and will be removed soon. Use the `runcard` argument instead."
        )
        mock_load.assert_called_once()
        mock_open.assert_called_once()

    def test_build_platform_passing_a_path_to_runcard_argument(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test build method."""
        platform = ql.build_platform(runcard="_")
        assert isinstance(platform, Platform)
        mock_load.assert_called_once()
        mock_open.assert_called_once()

    def test_build_platform_passing_a_dict_to_runcard_argument(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test build method."""
        platform = ql.build_platform(runcard=copy.deepcopy(Galadriel.runcard))
        assert isinstance(platform, Platform)
        mock_load.assert_not_called()
        mock_open.assert_not_called()

    def test_save_platform_with_non_yml_path(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test the `save_platform` function with a path that doesn't end in .yml."""
        path = save_platform(path="test/", platform=build_platform(Galadriel.runcard))

        mock_open.assert_called_once_with(file=Path(path), mode="w", encoding="utf-8")

        assert path == f"test/{Galadriel.name}.yml"
        mock_load.assert_not_called()

    def test_save_platform_with_yml_path(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test the `save_platform` function with a path that ends in .yml."""
        path = save_platform(path="test/test.yml", platform=build_platform(Galadriel.runcard))

        mock_open.assert_called_once_with(file=Path(path), mode="w", encoding="utf-8")

        assert path == "test/test.yml"
        mock_load.assert_not_called()


class TestBuildPlatformCornerCases:
    """Unit tests for the corner cases of the `build_platform` function.."""

    def test_build_method_with_no_arguments(self):
        """Test build method with the new drivers."""
        with pytest.raises(ValueError) as no_arg_error:
            _ = ql.build_platform()

            (msg,) = no_arg_error.value.args
            assert msg == "`runcard` argument (str | dict) has not been passed to the `build_platform()` function."

    def test_build_method_with_old_path_and_new_runcard_arguments(self):
        """Test build method with the new drivers."""
        with pytest.raises(
            ValueError,
            match="Use only the `runcard` argument, `path` argument is deprecated.",
        ):
            _ = ql.build_platform(runcard="_", path="_")

    def test_build_method_with_new_drivers(self):
        """Test build method with the new drivers."""
        with pytest.raises(NotImplementedError, match="New drivers are not supported yet"):
            _ = ql.build_platform(runcard="_", new_drivers=True)

    def test_platform_serialization_from_imported_dict(self):
        """Test platform serialization by building a platform, saving it and then load it back again twice. Starting from a given dict."""
        original_platform = ql.build_platform(Galadriel.runcard)
        path = save_platform(path="./test.yml", platform=original_platform)
        saved_platform = ql.build_platform(path)
        new_path = save_platform(path="./test.yml", platform=saved_platform)
        new_saved_platform = ql.build_platform(new_path)

        with open(file="examples/runcards/galadriel.yml", mode="r", encoding="utf8") as yaml_f:
            yaml_f_dict = yaml.safe_load(stream=yaml_f)
        with open(file="./test.yml", mode="r", encoding="utf8") as generated_f:
            generated_f_dict = yaml.safe_load(stream=generated_f)

        assert yaml_f_dict != generated_f_dict == Galadriel.runcard
        assert (
            original_platform.to_dict() == saved_platform.to_dict() == new_saved_platform.to_dict() == generated_f_dict
        )
        os.remove(path)  # Cleaning generated file

    def test_platform_serialization_from_yaml_file(self):
        """Test platform serialization by building a platform, saving it and then load it back again twice. Starting from a yaml file."""
        original_platform = ql.build_platform("examples/runcards/galadriel.yml")
        path = save_platform(path="./test.yml", platform=original_platform)
        saved_platform = ql.build_platform(path)
        new_path = save_platform(path="./test.yml", platform=saved_platform)
        new_saved_platform = ql.build_platform(new_path)

        with open(file="examples/runcards/galadriel.yml", mode="r", encoding="utf8") as yaml_f:
            yaml_f_dict = yaml.safe_load(stream=yaml_f)
        with open(file="./test.yml", mode="r", encoding="utf8") as generated_f:
            generated_f_dict = yaml.safe_load(stream=generated_f)

        for i in ["name", "device_id", "chip", "instruments", "instrument_controllers"]:
            assert yaml_f_dict[i] == generated_f_dict[i]
        assert generated_f_dict != Galadriel.runcard
        assert (
            original_platform.to_dict() == saved_platform.to_dict() == new_saved_platform.to_dict() == generated_f_dict
        )
        os.remove(path)  # Cleaning generated file


@patch("qililab.data_management.os.makedirs")
@patch("qililab.data_management.h5py.File")
class TestResultsData:
    """Unit tests for the `save_results` function."""

    def test_save_results_path(self, mock_file: MagicMock, mock_makedirs: MagicMock):
        """Test that the path created to save the results is correct."""

        path = save_results(results=np.array([]), loops={"one_loop": np.arange(0, 100, 0.1)}, data_path="", name="test")

        mock_file.assert_called_once_with(f"{path}/results.h5", "w")
        assert mock_makedirs.call_count == 2
        mock_makedirs.assert_called_with(path)

        assert path.endswith("_test")

    def test_load_results(self, mock_file: MagicMock, mock_makedirs: MagicMock):
        """Test that the path created to save the results is correct."""

        path = "this_is_a_test/results.h5"
        _ = load_results(path=path)

        mock_file.assert_called_once_with(path, "r")
        mock_makedirs.assert_not_called()



# pylint: disable=cyclic-import
# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains a pre-defined version of a T1 experiment."""
import numpy as np
from qibo.gates import M
from qibo.models import Circuit

import qililab as ql
from qililab.platform import Platform
from qililab.typings import ExperimentOptions, ExperimentSettings, Parameter
from qililab.utils import Loop, Wait

from .experiment_analysis import ExperimentAnalysis
from .fitting_models import Exp


class T1(ExperimentAnalysis, Exp):
    """Experiment to determine the T1 value of a qubit.

    The value of T1 is measured through the following sequence of operations:
        1. Prepare the qubit in the excited state by sending a π-pulse to the qubit.
        2. Wait some time t.
        3. Measure the state of the qubit.

    Args:
        platform (Platform): platform used to run the experiment
        qubit (int): qubit index used in the experiment
        wait_loop_values (numpy.ndarray): array of values to loop through in the experiment, which modifies the t parameter of the Wait gate
        repetition_duration (int, optional): duration of a single repetition in nanoseconds. Defaults to 200000.
        hardware_average (int, optional): number of repetitions used to average the result. Defaults to 10000.
    """

    def __init__(
        self,
        qubit: int,
        platform: Platform,
        wait_loop_values: np.ndarray,
        repetition_duration=200000,
        hardware_average=10000,
    ):
        # Define circuit used in this experiment
        circuit = Circuit(qubit + 1)
        circuit.add(ql.Drag(q=qubit, theta=np.pi, phase=0))
        circuit.add(Wait(q=qubit, t=0))
        circuit.add(M(qubit))

        # Define loop over wait duration
        loop = Loop(alias="2", parameter=Parameter.GATE_PARAMETER, values=wait_loop_values)

        # Define experiment options
        experiment_options = ExperimentOptions(
            name="t1",
            loops=[loop],
            settings=ExperimentSettings(repetition_duration=repetition_duration, hardware_average=hardware_average),
        )

        # Initialize experiment
        super().__init__(
            platform=platform,
            circuits=[circuit],
            options=experiment_options,
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utilities for Loops."""
import numpy as np

from qililab.utils.loop import Loop


def _find_minimum_range_from_loops(loops: list[Loop] | None):
    """find minimum range from same level loops"""
    if loops is None or len(loops) <= 0:
        return np.array([], dtype=object)
    minimum_range = loops[0].values
    minimum_range_length = len(loops[0].values)
    for loop in loops:
        if len(loop.values) < minimum_range_length:
            minimum_range = loop.values
            minimum_range_length = len(loop.values)
    return minimum_range


def _create_loops_from_inner_loops(loops: list[Loop]):
    """create sequence of loops from inner loops (if exist)"""
    return list(filter(None, [loop.loop for loop in loops]))


def compute_ranges_from_loops(loops: list[Loop] | None):
    """compute ranges from a list of loops that may have inner loops"""
    if loops is None or len(loops) <= 0:
        return []
    ranges = compute_ranges_from_loops(loops=_create_loops_from_inner_loops(loops=loops))
    ranges.append(_find_minimum_range_from_loops(loops=loops))
    return ranges


def compute_shapes_from_loops(loops: list[Loop]):
    """Computes the shape of the results obtained from running a list of parallel loops that might contain
    inner loops.

    When running parallel loops, the shape of the results correspond to the minimum range of each nested loop.

    Args:
        loops (list[Loop]): list of parallel loops that might contain inner loops

    Returns:
        list[int]: shape of the results obtained from running the parallel loops
    """
    if loops is None:
        return []
    all_shapes = [loop.shape for loop in loops]
    max_len = max(len(shape) for shape in all_shapes)
    final_shape: list[None | int] = [None] * max_len
    for shape in all_shapes:
        for i, dim in enumerate(shape):
            if final_shape[i] is None or dim < final_shape[i]:  # type: ignore
                final_shape[i] = dim
    return final_shape



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Circuit class

This file provides the Circuit class for representing quantum circuits.

It stores the circuit as a Directed Acyclic Graph and uses the rustworkx library for its manipulation.
It offers methods to add operations to the circuit, calculate the circuit's depth, and visualize the circuit.
"""
import rustworkx as rx
from rustworkx.visualization import graphviz_draw

from qililab.circuit.nodes import EntryNode, Node, OperationNode
from qililab.circuit.operations import Operation
from qililab.typings.enums import OperationTimingsCalculationMethod, Qubits


class Circuit:
    """The Circuit class

    Args:
        num_qubits (int): The number of qubits of the circuit

    Raises:
        ValueError: If num_qubits is not integer.
        ValueError: If num_qubits is not positive.
    """

    def __init__(self, num_qubits: int):
        if not isinstance(num_qubits, int):
            raise ValueError("Number of qubits should be integer.")

        if num_qubits <= 0:
            raise ValueError("Number of qubits should be positive.")

        self.num_qubits: int = num_qubits
        self.graph: rx.PyDiGraph = rx.PyDiGraph(multigraph=True)  # pylint: disable=no-member
        self.has_timings_calculated: bool = False
        self.has_special_operations_removed: bool = False
        self.has_transpiled_to_pulses: bool = False

        index = self.graph.add_node(EntryNode())
        self.graph[index].index = index
        self.entry_node = self.graph[index]

    def add(self, qubits: int | tuple[int, ...], operation: Operation, alias: str | None = None):
        """Adds an operation to the circuit.

        Args:
            qubits (int | tuple[int, ...]): The qubit(s) the operation acts on
            operation (Operation): The operation to add
            alias (str | None, optional): Optional alias for the operation. Defaults to None.
        """
        self._reset_transpilation_flags()
        qubits = qubits if isinstance(qubits, tuple) else (qubits,)
        if (operation.num_qubits == Qubits.ONE and len(qubits) != 1) or (
            operation.num_qubits == Qubits.TWO and len(qubits) != 2
        ):
            raise ValueError("Number of qubits does not match operation's num_qubits attribute")
        self._add_operation(qubits=qubits, operation=operation, alias=alias)

    def _add_operation(self, qubits: tuple[int, ...], operation: Operation, alias: str | None = None):
        """Adds one operation node for all qubits

        Args:
            qubits (tuple[int, ...]): The qubits the operation acts on
            operation (Operation): The operation to add
            alias (str | None, optional): Optional alias for the operation. Defaults to None.
        """
        new_operation_node = self._add_operation_node(qubits=qubits, operation=operation, alias=alias)
        last_operation_nodes = []
        for qubit in qubits:
            _, last_operation_node = self._last_operation_of_qubit(qubit=qubit)
            if last_operation_node not in last_operation_nodes:
                last_operation_nodes.append(last_operation_node)
        if any((isinstance(node, OperationNode) for node in last_operation_nodes)):
            for node in filter(lambda x: isinstance(x, OperationNode), last_operation_nodes):
                self.graph.add_edge(node.index, new_operation_node.index, None)
        else:
            self.graph.add_edge(self.entry_node.index, new_operation_node.index, None)

    def _add_operation_node(
        self, qubits: tuple[int, ...], operation: Operation, alias: str | None = None
    ) -> OperationNode:
        """Add an operation node to circuit's graph

        Args:
            qubits (tuple[int]): Tuple of qubits indices
            operation (Operation): The operation
            alias (str | None): Optional alias

        Returns:
            OperationNode: The operation node added
        """
        index = self.graph.add_node(OperationNode(operation=operation, qubits=qubits, alias=alias))
        self.graph[index].index = index
        return self.graph[index]

    def _last_operation_of_qubit(self, qubit: int) -> tuple[int, Node]:
        """Get the last operation node regarding qubit, along with the layer's index. If no operation found returns the entry node.

        Args:
            qubit (int): qubit's index

        Returns:
            tuple[int, Node]: layer index, Node
        """
        layers = rx.layers(self.graph, [self.entry_node.index])  # pylint: disable=no-member
        for index, layer in reversed(list(enumerate(layers[1:]))):
            for operation in layer:
                if qubit in operation.qubits:
                    return index, operation
        return 0, self.entry_node

    def _reset_transpilation_flags(self) -> None:
        """Reset the flags used for transpilation process"""
        self.has_timings_calculated = False
        self.has_special_operations_removed = False
        self.has_transpiled_to_pulses = False

    @property
    def depth(self) -> int:
        """Get the depth of the circuit which is essentially the number of operation layers

        Returns:
            int: The depth of the circuit
        """
        return len(rx.layers(self.graph, [self.entry_node.index])) - 1  # pylint: disable=no-member

    def get_operation_layers(
        self, method: OperationTimingsCalculationMethod = OperationTimingsCalculationMethod.AS_SOON_AS_POSSIBLE
    ) -> list[list[OperationNode]]:
        """Get the layers of operation nodes. Each layer represents an advancement in time.

        Args:
            method (OperationTimingsCalculationMethod, optional): The method that layers should be calculated.
                If set to `OperationTimingsCalculationMethod.AS_LATE_AS_POSSIBLE`, we rearrange the layers, moving
                operations to the largest layer index possible. Defaults to
                OperationTimingsCalculationMethod.AS_SOON_AS_POSSIBLE.

        Returns:
            list[list[OperationNode]]: A list of layers each containing a list of operation nodes.
                Operation nodes are sorted based on their index. (order of insertion)
        """
        layers = rx.layers(self.graph, [self.entry_node.index])[1:]  # pylint: disable=no-member
        for layer in layers:
            layer.sort(key=lambda node: node.index)
        if method == OperationTimingsCalculationMethod.AS_SOON_AS_POSSIBLE:
            return layers
        for qubit in range(self.num_qubits):
            for index, layer in enumerate(layers):
                current_single_operations_on_qubit = [
                    op_i
                    for op_i, operation in enumerate(layer)
                    if qubit in operation.qubits and len(operation.qubits) == 1
                ]
                if len(current_single_operations_on_qubit) == 0:
                    continue
                if index == len(layers) - 1:
                    continue
                next_layer_operations_on_qubit = [
                    operation for operation in layers[index + 1] if qubit in operation.qubits
                ]
                if len(next_layer_operations_on_qubit) == 0:
                    layers[index + 1].append(layer.pop(current_single_operations_on_qubit[0]))
        return layers

    def draw(self, filename: str | None = None):
        """Draws the circuit's graph.

        Args:
            filename (str | None, optional): If set, the result is saved to a file. Defaults to None.
        """

        def node_attr(node):
            if isinstance(node, EntryNode):
                return {"color": "yellow", "fillcolor": "yellow", "style": "filled", "label": "start"}
            operation = str(node.operation)
            qubits = ", ".join((str(qubit) for qubit in node.qubits))
            timing = f"{node.timing.start}ns -> {node.timing.end}ns" if node.timing is not None else ""
            label = f"{operation}: {qubits}\n{timing}"
            return {"color": "red", "fillcolor": "red", "style": "filled", "label": label}

        image = graphviz_draw(self.graph, node_attr_fn=node_attr, filename=filename)
        if image is not None:
            image.show()

    def print(self, method: OperationTimingsCalculationMethod = OperationTimingsCalculationMethod.AS_SOON_AS_POSSIBLE):
        """Prints the circuit to the standard output.

        Args:
            method (OperationTimingsCalculationMethod, optional): Defaults to OperationTimingsCalculationMethod.AS_SOON_AS_POSSIBLE.
        """
        layers = self.get_operation_layers(method=method)
        for qubit in range(self.num_qubits):
            print(f"{qubit}:", end="")
            for _, layer in enumerate(layers):
                operations_on_qubit = [operation for operation in layer if qubit in operation.qubits]
                if len(operations_on_qubit) == 0:
                    print("----------", end="")
                if len(operations_on_qubit) == 1:
                    print(f"{operations_on_qubit[0].operation.name.value:->10s}", end="")
            print()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the functions used to decompose a circuit into native gates and to compute virtual-Z gates."""
from qibo import gates
from qibo.models import Circuit

from qililab.settings.runcard import Runcard
from qililab.transpiler.gate_decompositions import translate_gates

from .native_gates import Drag


def translate_circuit(
    circuit: Circuit, gates_settings: Runcard.GatesSettings | None = None, optimize: bool = True
) -> Circuit:
    """Converts circuit with qibo gates to circuit with native gates

    Args:
        circuit (Circuit): circuit with qibo gates
        optimize (bool): optimize the transpiled circuit using otpimize_transpilation
        gates_settings (Runcard.GatesSettings): calibrated gate settings from the runcard. This is used only for phase correction in CZ gates

    Returns:
        new_circuit (Circuit): circuit with transpiled gates
    """
    # get list of gates from circuit
    ngates = circuit.queue
    # init new circuit
    new_circuit = Circuit(circuit.nqubits)
    # add transpiled gates to new circuit, optimize if needed
    if optimize:
        gates_to_optimize = translate_gates(ngates)
        new_circuit.add(
            optimize_transpilation(circuit.nqubits, ngates=gates_to_optimize, gates_settings=gates_settings)
        )
    else:
        new_circuit.add(translate_gates(ngates))

    return new_circuit


def optimize_transpilation(
    nqubits: int, ngates: list[gates.Gate], gates_settings: Runcard.GatesSettings | None = None
) -> list[gates.Gate]:
    """Optimizes transpiled circuit by applying virtual Z gates.
    This is done by moving all RZ to the left of all operators as a single RZ. The corresponding cumulative rotation
    from each RZ is carried on as phase in all drag pulses left of the RZ operator.
    Virtual Z gates are also applied to correct phase errors from CZ gates.
    The final RZ operator left to be applied as the last operator in the circuit can afterwards be removed since the last
    operation is going to be a measurement, which is performed on the Z basis and is therefore invariant under rotations
    around the Z axis.
    This last step can also be seen from the fact that an RZ operator applied on a single qubit, with no operations carried
    on afterwards induces a phase rotation. Since phase is an imaginary unitary component, its absolute value will be 1
    independent on any (unitary) operations carried on it.

    Mind that moving an operator to the left is equivalent to applying this operator last so
    it is actually moved to the _right_ of Circuit.queue (last element of list).

    For more information on virtual Z gates, see https://arxiv.org/abs/1612.00858

    Args:
        nqubits (int) : number of qubits in the circuit
        ngates (list[gates.Gate]) : list of gates in the circuit
        gates_settings (Runcard.GatesSettings): calibrated gate settings from the runcard. This is used only for phase correction in CZ gates

    Returns:
        list[gates.Gate] : list of re-ordered gates
    """
    supported_gates = ["rz", "drag", "cz", "measure"]
    new_gates = []
    shift = {qubit: 0 for qubit in range(nqubits)}
    for gate in ngates:
        if gate.name not in supported_gates:
            raise NotImplementedError(f"{gate.name} not part of supported gates {supported_gates}")
        if isinstance(gate, gates.RZ):
            shift[gate.qubits[0]] += gate.parameters[0]
        # add CZ phase correction
        elif isinstance(gate, gates.CZ) and gates_settings is not None:
            gate_settings = gates_settings.get_gate(name=gate.__class__.__name__, qubits=gate.qubits)
            control_qubit, target_qubit = gate.qubits
            corrections = next(
                (
                    event.pulse.options
                    for event in gate_settings
                    if (event.pulse.options is not None and f"q{control_qubit}_phase_correction" in event.pulse.options)
                ),
                None,
            )
            if corrections is not None:
                shift[control_qubit] += corrections[f"q{control_qubit}_phase_correction"]
                shift[target_qubit] += corrections[f"q{target_qubit}_phase_correction"]
            new_gates.append(gate)
        else:
            # if gate is drag pulse, shift parameters by accumulated Zs
            if isinstance(gate, Drag):
                # create new drag pulse rather than modify parameters of the old one
                gate = Drag(gate.qubits[0], gate.parameters[0], gate.parameters[1] - shift[gate.qubits[0]])

            # append gate to optimized list
            new_gates.append(gate)

    return new_gates



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Yaml utils."""
import yaml


def yaml_representer(dumper: yaml.Dumper, value: int | float):
    """Int or float representer used by YAML."""
    tag = "int" if isinstance(value, int) else "float"
    text = f"{value:g}"
    if "e" in text:
        tag = "float"
        if "." not in text.split("e")[0]:
            # Add a decimal point if value is integer
            text = text.split("e")[0] + ".e" + "".join(text.split("e")[1:])
    if "." not in text and tag == "float":
        # Change tag to int if there is no decimal point.
        tag = "int"
    return dumper.represent_scalar(tag=f"tag:yaml.org,2002:{tag}", value=text)


def null_representer(dumper: yaml.Dumper, value: None):  # pylint: disable=unused-argument
    """Int or float representer used by YAML."""
    return dumper.represent_scalar(tag="tag:yaml.org,2002:null", value="")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from qililab.circuit.operation_factory import OperationFactory
from qililab.circuit.operations.special_operations.special_operation import SpecialOperation
from qililab.typings.enums import OperationName, Qubits
from qililab.utils import classproperty


@OperationFactory.register
@dataclass
class Reset(SpecialOperation):
    """Operation representing a reset to ground state."""

    @classproperty
    def name(self) -> OperationName:
        """Get operation's name

        Returns:
            OperationName: The operation's name
        """
        return OperationName.RESET

    @classproperty
    def num_qubits(self) -> Qubits:
        """Get number of qubits the operation can act upon

        Returns:
            Qubits: The number of qubits the operation can act upon
        """
        return Qubits.ANY



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CircuitTranspiler class."""
from copy import deepcopy
from dataclasses import dataclass

from qililab.settings import Runcard
from qililab.typings.enums import ResetMethod

from .circuit import Circuit
from .nodes.operation_node import OperationTiming
from .operation_factory import OperationFactory
from .operations import Barrier, Measure, PulseOperation, Reset, SpecialOperation, TranslatableToPulseOperation, Wait


@dataclass
class CircuitTranspiler:
    """CircuitTranspiler class."""

    settings: Runcard.GatesSettings

    def calculate_timings(self, circuit: Circuit) -> Circuit:
        """Calculates the timings of all operations in a given quantum circuit and annotates the operation nodes with timing information.

        Args:
            circuit (Circuit): The quantum circuit for which timings need to be calculated.
        Returns:
            Circuit: The circuit with annotated operations that contain timing information.

        Raises:
            ValueError: If an unsupported operation is encountered in the circuit.
        """
        circuit = deepcopy(circuit)
        nqubits = circuit.num_qubits
        layers = circuit.get_operation_layers(method=self.settings.timings_calculation_method)
        qubits_last_end_timings = [0 for _ in range(nqubits)]
        for index, layer in enumerate(layers):
            # Calculate maximum end time of previous layer
            max_end_time_of_previous_layer = (
                max(op_node.timing.end for op_node in layers[index - 1] if op_node.timing is not None)
                if index >= 1
                else 0
            )
            for operation_node in layer:
                # Initial guess of start time of operation
                start_time = max(
                    [qubits_last_end_timings[qubit] for qubit in operation_node.qubits]
                    + [max_end_time_of_previous_layer]
                )
                # Calculate [start-end] time of operation
                if isinstance(operation_node.operation, TranslatableToPulseOperation):
                    operation_settings = self.settings.get_operation_settings(operation_node.operation.name.value)
                    pulse_duration = operation_settings.pulse.duration
                    delay = (
                        self.settings.delay_before_readout
                        if isinstance(operation_node.operation, Measure)
                        else self.settings.delay_between_pulses
                    )
                    start_time = start_time + delay
                    end_time = start_time + pulse_duration
                elif isinstance(operation_node.operation, PulseOperation):
                    start_time = start_time + self.settings.delay_between_pulses
                    end_time = start_time + operation_node.operation.duration
                elif isinstance(operation_node.operation, Wait):
                    end_time = start_time + operation_node.operation.t
                elif isinstance(operation_node.operation, Barrier):
                    end_time = start_time
                elif isinstance(operation_node.operation, Reset):
                    if self.settings.reset_method == ResetMethod.PASSIVE:
                        end_time = start_time + self.settings.passive_reset_duration
                else:
                    raise ValueError(f"Operation {operation_node} not supported for translation yet.")
                # Update timings
                operation_node.timing = OperationTiming(start=start_time, end=end_time)
                for qubit in operation_node.qubits:
                    qubits_last_end_timings[qubit] = end_time
        circuit.has_timings_calculated = True
        return circuit

    def remove_special_operations(self, circuit: Circuit) -> Circuit:
        """Removes special operations (Wait, Barrier, Passive Reset) from the circuit

        Args:
            circuit (Circuit): The quantum circuit

        Returns:
            Circuit: New circuit with special operations removed
        """
        circuit = deepcopy(circuit)
        if not circuit.has_timings_calculated:
            circuit = self.calculate_timings(circuit)
        layers = circuit.get_operation_layers(method=self.settings.timings_calculation_method)
        for layer in layers:  # pylint: disable=too-many-nested-blocks
            for operation_node in layer:
                if isinstance(operation_node.operation, SpecialOperation):
                    predecessors = circuit.graph.predecessors(operation_node.index)
                    successors = circuit.graph.successors(operation_node.index)
                    for predecessor in predecessors:
                        for successor in successors:
                            predecessor_qubits = set(predecessor.qubits)
                            successor_qubits = set(successor.qubits)
                            node_qubits = set(operation_node.qubits)
                            if predecessor_qubits & successor_qubits & node_qubits:
                                circuit.graph.add_edge(predecessor.index, successor.index, None)
                    circuit.graph.remove_node(operation_node.index)
        circuit.has_special_operations_removed = True
        return circuit

    def transpile_to_pulse_operations(self, circuit: Circuit) -> Circuit:
        """
        Transpiles the given quantum circuit into pulse operations.

        Args:
            circuit (Circuit): The quantum circuit to be transpiled into pulse operations.

        Returns:
            Circuit: The transpiled circuit with pulse operations.
        """
        circuit = deepcopy(circuit)
        if not circuit.has_timings_calculated:
            circuit = self.calculate_timings(circuit)
        if not circuit.has_special_operations_removed:
            circuit = self.remove_special_operations(circuit)
        layers = circuit.get_operation_layers(method=self.settings.timings_calculation_method)
        for layer in layers:
            for operation_node in layer:
                if isinstance(operation_node.operation, TranslatableToPulseOperation):
                    operation_settings = self.settings.get_operation_settings(operation_node.operation.name.value)
                    pulse_operation_settings = operation_settings.pulse
                    pulse_operation_name = pulse_operation_settings.name
                    pulse_operation_parameters = {
                        "amplitude": pulse_operation_settings.amplitude,
                        "duration": pulse_operation_settings.duration,
                        **pulse_operation_settings.parameters,
                    }
                    pulse_operation = OperationFactory.get(pulse_operation_name)(**pulse_operation_parameters)
                    operation_node.operation = pulse_operation
        circuit.has_transpiled_to_pulses = True
        return circuit



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""__init__.py"""
from .enums import (
    AcquireTriggerMode,
    AcquisitionName,
    ConnectionName,
    GateName,
    Instrument,
    InstrumentName,
    IntegrationMode,
    NodeName,
    OperationName,
    Parameter,
    PulseDistortionName,
    PulseShapeName,
    ReferenceClock,
    ResultName,
)
from .experiment import ExperimentOptions, ExperimentSettings
from .factory_element import FactoryElement
from .instruments import (
    Cluster,
    Device,
    Keithley2600Driver,
    MiniCircuitsDriver,
    Pulsar,
    QbloxD5a,
    QbloxS4g,
    QcmQrm,
    RohdeSchwarzSGS100A,
)
from .yaml_type import yaml



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ExecutionBuilder class"""
from warnings import warn

from qililab.platform import Platform
from qililab.pulse import PulseSchedule
from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
from qililab.utils import Loop, Singleton

from .bus_execution import BusExecution
from .execution_manager import ExecutionManager


class ExecutionBuilder(metaclass=Singleton):
    """Builder of platform objects."""

    def build(self, platform: Platform, pulse_schedules: list[PulseSchedule]) -> ExecutionManager:
        """Build ExecutionManager class.
        Loop over pulses in PulseSequence, classify them by bus index and instantiate a BusExecution class.

        Returns:
            ExecutionManager: ExecutionManager object.
        """
        buses: dict[int, BusExecution] = {}
        for pulse_schedule in pulse_schedules:
            for pulse_bus_schedule in pulse_schedule.elements:
                port, bus = self._get_bus_info_from_pulse_bus_schedule_port(platform, pulse_bus_schedule)
                if bus is None:
                    raise ValueError(f"There is no bus connected to port {port}.")
                if port not in buses:
                    buses[port] = BusExecution(bus=bus, pulse_bus_schedules=[pulse_bus_schedule])
                    continue
                buses[port].add_pulse_bus_schedule(pulse_bus_schedule=pulse_bus_schedule)

        return ExecutionManager(buses=list(buses.values()), num_schedules=len(pulse_schedules))

    def build_from_loops(self, platform: Platform, loops: list[Loop]) -> ExecutionManager:
        """Build ExecutionManager class.
        Loop over loops, classify them by bus alias and instantiate a BusExecution class.

        Returns:
            ExecutionManager: ExecutionManager object.
        """
        warn(
            "|WARNING| Bus alias are not unique and can be repeated in the runcard\nThe first bus alias that matches the loop alias will be selected"
        )
        buses: dict[str, BusExecution] = {}
        for loop in loops:
            for _loop in loop.loops:  # Iterate over nested loops if any
                alias, bus = self._get_bus_info_from_loop_alias(platform, _loop)
                if bus is None:
                    raise ValueError(
                        f"There is no bus with alias '{alias}'\n|INFO| Make sure the loop alias matches the bus alias specified in the runcard"
                    )
                if alias in buses:
                    warn(
                        f"|WARNING| Loop alias is repeated\nBus execution for bus with alias '{alias}' already created, skipping iteration"
                    )
                else:
                    buses[alias] = BusExecution(bus=bus, pulse_bus_schedules=[])

        return ExecutionManager(buses=list(buses.values()), num_schedules=0)

    def _get_bus_info_from_pulse_bus_schedule_port(self, platform: Platform, pulse_bus_schedule: PulseBusSchedule):
        """get the bus information that it is connected to the port in the pulse bus schedule
        Args:
            platform: Platform
            pulse_bus_schedule: PulseBusSchedule
        Returns:
            port: pulse_bus_schedule.port
            bus_idx: index of the bus
            bus: Bus object
        """
        port = pulse_bus_schedule.port
        bus = platform.buses.get(port=port)
        return port, bus

    def _get_bus_info_from_loop_alias(self, platform: Platform, loop: Loop):
        """get the bus information that it is connected to the port from the loop alias. Loop alias has to be the same as the bus alias
        Args:
            platform: Platform
            loop: Loop
        Returns:
            alias: alias of the bus
            bus: Bus object
        """
        alias = loop.alias
        bus = platform._get_bus_by_alias(alias=alias)  # pylint: disable=protected-access
        return alias, bus



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

from qililab.pulse import PulseBusSchedule

from .base_instrument import BaseInstrument
from .instrument_interface_factory import InstrumentInterfaceFactory


@InstrumentInterfaceFactory.register
class AWG(BaseInstrument):
    """
    Interface for AWG sequencer instrument types.
    """

    @abstractmethod
    def execute(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ) -> None:
        """Compiles a pulse bus schedule, generates associated QASM program and runs it.

        Args:
            pulse_bus_schedule (PulseBusSchedule): Pulse Bus Schedule to generate QASM program.
            nshots (int): number of shots
            repetition_duration (int): repetition duration.
            num_bins (int): number of bins
        """



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Experiment Options Typings """

from dataclasses import asdict, dataclass, field

from qililab.constants import EXPERIMENT, RUNCARD
from qililab.typings.yaml_type import yaml
from qililab.utils.loop import Loop

DEFAULT_EXPERIMENT_NAME = "experiment"


@dataclass
class ExperimentSettings:
    """Experiment settings."""

    hardware_average: int = 1024
    software_average: int = 1
    repetition_duration: int = 200000
    num_bins: int = 1

    def __str__(self):
        """Returns a string representation of the experiment settings."""
        return yaml.dump(asdict(self), sort_keys=False)


@dataclass
class ExperimentOptions:
    """Experiment Options"""

    loops: list[Loop] | None = None
    settings: ExperimentSettings = field(default_factory=ExperimentSettings)
    name: str = DEFAULT_EXPERIMENT_NAME
    remote_save: bool = True
    description: str = ""

    def to_dict(self):
        """Convert Experiment into a dictionary.

        Returns:
            dict: Dictionary representation of the Experiment class.
        """
        return {
            EXPERIMENT.LOOPS: [loop.to_dict() for loop in self.loops] if self.loops is not None else None,
            RUNCARD.GATES_SETTINGS: asdict(self.settings),
            RUNCARD.NAME: self.name,
            EXPERIMENT.REMOTE_SAVE: self.remote_save,
            EXPERIMENT.DESCRIPTION: self.description,
        }

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Load Experiment Options from dictionary.

        Args:
            dictionary (dict): Dictionary description of an experiment.
        """
        input_loops = dictionary[EXPERIMENT.LOOPS]
        loops = [Loop(**loop) for loop in input_loops] if input_loops is not None else None

        return ExperimentOptions(
            loops=loops,
            settings=ExperimentSettings(**dictionary[RUNCARD.GATES_SETTINGS])
            if RUNCARD.GATES_SETTINGS in dictionary
            else ExperimentSettings(),
            name=dictionary[RUNCARD.NAME] if RUNCARD.NAME in dictionary else DEFAULT_EXPERIMENT_NAME,
            remote_save=dictionary.get(EXPERIMENT.REMOTE_SAVE, True),
            description=dictionary.get(EXPERIMENT.DESCRIPTION, ""),
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OperationFactory class"""
from typing import Callable, Type


class OperationFactory:
    """A factory class to register and retrieve operation classes based on their names."""

    _operations: dict[str, Type] = {}

    @classmethod
    def register(cls, operation: Type) -> Callable:
        """Registers an operation class to the factory.

        Args:
            operation (Type): The operation class to be registered.

        Returns:
            Callable: A decorator that registers the operation class to the factory.
        """

        def decorator(operation: Type) -> Type:
            cls._operations[operation.name.value] = operation
            return operation

        return decorator(operation=operation)

    @classmethod
    def get(cls, name: str) -> Type:
        """Retrieves the operation class registered with the given name.

        Args:
            name (str): The name of the operation class to retrieve.

        Returns:
            Type: The operation class registered with the given name.

        Raises:
            ValueError: If the operation class with the given name is not registered.
        """
        if name in cls._operations:
            return cls._operations[name]
        raise ValueError(f"Operation '{name}' is not registered.")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains a pre-defined version of the flux spectroscopy experiment."""
import numpy as np
from qibo.gates import M
from qibo.models import Circuit

import qililab as ql  # pylint: disable=cyclic-import
from qililab.platform import Platform
from qililab.typings import ExperimentOptions, ExperimentSettings, Parameter
from qililab.utils import Loop, Wait

from .experiment_analysis import ExperimentAnalysis
from .fitting_models import Exp


class T2Echo(ExperimentAnalysis, Exp):
    """Class used to run a T2Echo experiment on the given qubit. The experiment performs the following circuit:

    1. Prepare the qubit in the ∣-i⟩ state by sending a π/2-pulse to the qubit
    2. Wait some time t
    3. Send a π-pulse to the qubit
    4. Wait some time t
    5. Send a π/2-pulse to the qubit
    6. Measure the qubit

    Args:
        platform (Platform): platform used to run the experiment.
        qubit (int): qubit index used in the experiment.
        wait_loop_values (ndarray): array of time values to wait for the `Wait` gates, the same value is used on both
            gates.
        repetition_duration (int, optional): duration of a single repetition in nanoseconds. Default to 10000.
        measurement_buffer (int, optional): time to wait before taking the measurement. Default to 100.
        hardware_average (int, optional): number of repetitions used to average the result. Default to 10000.
    """

    def __init__(
        self,
        qubit: int,
        platform: Platform,
        wait_loop_values: np.ndarray,
        repetition_duration=10000,
        measurement_buffer=100,
        hardware_average=10000,
    ):
        self.wait_loop_values = wait_loop_values
        self.t2 = None
        # Define Circuit to execute
        circuit = Circuit(qubit + 1)
        circuit.add(ql.Drag(q=qubit, theta=np.pi / 2, phase=0))
        circuit.add(Wait(qubit, t=0))
        circuit.add(ql.Drag(q=qubit, theta=np.pi, phase=0))
        circuit.add(Wait(qubit, t=0))
        circuit.add(ql.Drag(q=qubit, theta=np.pi / 2, phase=0))
        circuit.add(Wait(qubit, t=measurement_buffer))
        circuit.add(M(qubit))

        # Alias of loops reference the wait parameters, looping over wait times.
        first_wait_loop = Loop(alias="2", parameter=Parameter.GATE_PARAMETER, values=wait_loop_values)
        second_wait_loop = Loop(alias="5", parameter=Parameter.GATE_PARAMETER, values=wait_loop_values)

        experiment_options = ExperimentOptions(
            name="t2echo",
            loops=[first_wait_loop, second_wait_loop],
            settings=ExperimentSettings(repetition_duration=repetition_duration, hardware_average=hardware_average),
        )

        super().__init__(
            platform=platform,
            circuits=[circuit],
            options=experiment_options,
        )

    def fit(self, p0: tuple | None = (-52, 2000, 0)):
        """
        Fitting function for the T2Echo class. Calls the ExperimentAnalysis.fit() and returns the fitted T2 value.

        Args:
            p0 (tuple, optional): Initial guess for the parameters. Default to (-52, 2000, 0).
        """
        fitted_params = super().fit(p0=p0)
        self.t2 = 2 * fitted_params[1]
        return self.t2



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QubitControl class."""
from abc import abstractmethod
from dataclasses import asdict, dataclass
from typing import Sequence

from qililab.constants import RUNCARD
from qililab.instruments.awg_settings.awg_sequencer import AWGSequencer
from qililab.instruments.awg_settings.typings import AWGTypes
from qililab.instruments.instrument import Instrument
from qililab.pulse import PulseBusSchedule
from qililab.utils.asdict_factory import dict_factory


class AWG(Instrument):
    """Abstract base class defining all instruments used to control or readout the qubits."""

    @dataclass(kw_only=True)
    class AWGSettings(Instrument.InstrumentSettings):
        """Contains the settings of a AWG.

        Args:
            num_sequencers (int): Number of sequencers (physical I/Q pairs)
            awg_sequencers (Sequence[AWGSequencer]): Properties of each AWG sequencer
        """

        num_sequencers: int
        awg_sequencers: Sequence[AWGSequencer]

        def __post_init__(self):
            """build AWGSequencers and IQ channels"""
            super().__post_init__()
            if self.num_sequencers <= 0:
                raise ValueError(f"The number of sequencers must be greater than 0. Received: {self.num_sequencers}")
            if len(self.awg_sequencers) != self.num_sequencers:
                raise ValueError(
                    f"The number of sequencers: {self.num_sequencers} does not match"
                    + f" the number of AWG Sequencers settings specified: {len(self.awg_sequencers)}"
                )
            self.awg_sequencers = [
                AWGSequencer(**sequencer) if isinstance(sequencer, dict) else sequencer  # pylint: disable=not-a-mapping
                for sequencer in self.awg_sequencers
            ]

        def to_dict(self):
            """Return a dict representation of an AWG instrument."""
            result = asdict(self, dict_factory=dict_factory)
            result.pop(AWGTypes.AWG_SEQUENCERS.value)

            return result | {AWGTypes.AWG_SEQUENCERS.value: [sequencer.to_dict() for sequencer in self.awg_sequencers]}

    settings: AWGSettings

    @abstractmethod
    def compile(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ) -> list:
        """Compiles the ``PulseBusSchedule`` into an assembly program.

        Args:
            pulse_bus_schedule (PulseBusSchedule): the list of pulses to be converted into a program
            nshots (int): number of shots / hardware average
            repetition_duration (int): repetition duration
            num_bins (int): number of bins

        Returns:
            list: list of compiled assembly programs
        """

    @abstractmethod
    def run(self, port: str):
        """Run the uploaded program"""

    @abstractmethod
    def upload(self, port: str):
        """Upload compiled program."""

    @property
    def num_sequencers(self):
        """Number of sequencers in the AWG

        Returns:
            int: number of sequencers
        """
        return self.settings.num_sequencers

    @property
    def awg_sequencers(self):
        """AWG 'awg_sequencers' property."""
        return self.settings.awg_sequencers

    @property
    def intermediate_frequencies(self):
        """AWG 'intermediate_frequencies' property."""
        return [sequencer.intermediate_frequency for sequencer in self.awg_sequencers]

    def to_dict(self):
        """Return a dict representation of an AWG instrument."""
        return {RUNCARD.NAME: self.name.value} | self.settings.to_dict()

    def get_sequencers_from_chip_port_id(self, chip_port_id: str):
        """Get sequencer ids from the chip port identifier

        Args:
            chip_port_id (str): chip port identifier

        Returns:
            list[AWGSequencer]: list of integers containing the indices of the sequencers connected to the chip port
        """
        if seqs := [sequencer for sequencer in self.awg_sequencers if sequencer.chip_port_id == chip_port_id]:
            return seqs
        raise IndexError(
            f"No sequencer found connected to port {chip_port_id}. Please make sure the `chip_port_id` "
            "attribute is correct."
        )

    def get_sequencer(self, sequencer_id: int) -> AWGSequencer:
        """Get sequencer from the sequencer identifier

        Args:
            sequencer_id (int): sequencer identifier

        Returns:
            AWGSequencer: sequencer associated with the sequencer_id
        """
        sequencer_identifiers = [
            sequencer.identifier for sequencer in self.awg_sequencers if sequencer.identifier == sequencer_id
        ]

        if len(sequencer_identifiers) != 1:
            raise ValueError(
                f"Each sequencer should have a unique id. Found {len(sequencer_identifiers)} sequencers "
                f"with id {sequencer_id}."
            )

        return self.awg_sequencers[sequencer_identifiers[0]]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OperationNode class"""
from dataclasses import dataclass

from qililab.circuit.nodes.node import Node
from qililab.circuit.operations import Operation


@dataclass
class OperationTiming:
    """Class to store timings of an operation"""

    start: int
    end: int


@dataclass
class OperationNode(Node):
    """Node representing an operation acting on one or more qubits

    Args:
        operation (Operation): The operation that is applied
        Node (tuple[int]): The qubits the operation is applied on
        alias (str | None): An optional alias for easily retrieving the node (default = None)
    """

    operation: Operation
    qubits: tuple[int, ...]
    alias: str | None = None
    timing: OperationTiming | None = None



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field

from qililab.circuit.operation_factory import OperationFactory
from qililab.circuit.operations.translatable_to_pulse_operations.translatable_to_pulse_operation import (
    TranslatableToPulseOperation,
)
from qililab.typings import OperationName
from qililab.typings.enums import Qubits
from qililab.utils import classproperty


@OperationFactory.register
@dataclass
class Rxy(TranslatableToPulseOperation):
    """Operation representing a rotation around XY axis

    Args:
        theta (float): theta angle in degrees
        phi (float): phi angle in degrees
    """

    theta: float
    phi: float

    @classproperty
    def name(self) -> OperationName:
        return OperationName.RXY

    @classproperty
    def num_qubits(self) -> Qubits:
        """Get number of qubits the operation can act upon

        Returns:
            Qubits: The number of qubits the operation can act upon
        """
        return Qubits.ONE

    @property
    def parameters(self):
        """Get the names and values of all parameters as dictionary

        Returns:
            Parameters: The parameters of the operation
        """
        return {"theta": self.theta, "phi": self.phi}


@OperationFactory.register
@dataclass
class R180(Rxy):
    """R180 Operation

    Operation representing a pi rotation around XY axis

    Args:
        phi (float): phi angle in degrees
    """

    theta: float = field(init=False, default=180)
    phi: float

    @classproperty
    def name(self) -> OperationName:
        """Get operation's name

        Returns:
            OperationName: The operation's name
        """
        return OperationName.R180

    @property
    def parameters(self):
        """Get the names and values of all parameters as dictionary

        Returns:
            Parameters: The parameters of the operation
        """
        return {"phi": self.phi}


@OperationFactory.register
@dataclass
class X(R180):
    """X Operation

    Operation representing a pi rotation around XY axis with zero phase. It is the equivelant to an X gate.
    """

    theta: float = field(init=False, default=180)
    phi: float = field(init=False, default=0)

    @classproperty
    def name(self) -> OperationName:
        """Get operation's name

        Returns:
            OperationName: The operation's name
        """
        return OperationName.X

    @property
    def parameters(self):
        """Get the names and values of all parameters as dictionary

        Returns:
            Parameters: The parameters of the operation
        """
        return {}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This function contains all the fitting models used in experiment analysis."""
from abc import ABC, abstractmethod

import numpy as np

# pylint: disable=too-few-public-methods


class FittingModel(ABC):
    """Abstract Base Class defining a fitting model.

    An experiment analysis should inherit from a child of this class, such that the ``func`` method can be used
    to fit the experiment results.

    A class is used instead of a function to make sure the user can see the signature of the fitting function.
    """

    @staticmethod
    @abstractmethod
    def func(xdata: np.ndarray, *args):
        """The model function, func(x, …) used to fit the post-processed data.

        It must take the independent variable as the first argument and the parameters to fit as separate remaining
        arguments.

        Args:
            xdata (np.ndarray): independent variable
            *args: parameters to fit

        Returns:
            np.ndarray: model function evaluated at xdata
        """


class Cos(FittingModel):
    """Cosine model function."""

    @staticmethod
    def func(xdata: np.ndarray, amplitude: float, frequency: float, phase: float, offset: float) -> np.ndarray:  # type: ignore  # pylint: disable=arguments-differ
        """Cosine model function.

        It must take the independent variable as the first argument and the parameters to fit as separate remaining
        arguments.

        Args:
            xdata (ndarray): amplitude of the X gate
            amplitude (float): amplitude of the cosine function
            frequency (float): frequency in Hz (f, not omega!)
            phase (float): phase in rad
            offset (float): offset
        """
        return amplitude * np.cos(2 * np.pi * frequency * xdata + phase) + offset


class Exp(FittingModel):
    """Exponential model function."""

    @staticmethod
    def func(xdata: np.ndarray, amplitude: float, rate: float, offset: float) -> np.ndarray:  # type: ignore  # pylint: disable=arguments-differ
        """Exponential model function.

        It must take the independent variable as the first argument and the parameters to fit as separate remaining
        arguments.

        Args:
            xdata (ndarray): data of the x-axis
            amplitude (float): amplitude of the exponential function
            rate (float): rate of growth or decay
            offset (float): offset
        """
        return amplitude * np.exp(rate * xdata) + offset



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the ``ExperimentAnalysis`` class used to analyze the results of an experiment."""
import inspect

import matplotlib.pyplot as plt
import numpy as np
from qibo.models import Circuit
from scipy.optimize import curve_fit

from qililab.experiment.experiment import Experiment
from qililab.platform import Bus, Platform
from qililab.typings import ExperimentOptions, Parameter
from qililab.utils import Loop

from .fitting_models import FittingModel


class ExperimentAnalysis(Experiment, FittingModel):
    """Class used to analyze the results of an experiment. The analysis of an experiment consist of the following steps:

    1. Acquire results: either by running an experiment or loading one.
    2. Post-process results: post-process the results of the experiment.
    3. Fit results: fit the post-processed results of the experiment.
    4. Plot results: plot the post-processed results of the experiment together with the fitted function.

    Args:
        platform (Platform): platform used to run the experiment
        circuits (list[Circuit]): list of circuits used in the experiment
        options (ExperimentOptions): options of the experiment
        control_bus (Bus, optional): control bus used in the experiment. Defaults to None.
        readout_bus (Bus, optional): readout bus used in the experiment. Defaults to None.
        experiment_loop (.Loop, optional): external loop used in the experiment. This argument can be used for
            experiments that use a loop to define multiple circuits, such as the Flipping Sequence experiment.
            Defaults to None.
    """

    post_processed_results: np.ndarray
    popt: np.ndarray  # fitted parameters

    def __init__(  # pylint: disable=too-many-arguments
        self,
        platform: Platform,
        circuits: list[Circuit],
        options: ExperimentOptions,
        control_bus: Bus | None = None,  # TODO: This will probably change for 2-qubit experiments
        readout_bus: Bus | None = None,
        experiment_loop: Loop | None = None,
    ):
        if experiment_loop is None:
            if options.loops is None:
                raise ValueError(
                    "A loop must be provided. Either an experiment loop in the `ExperimentOptions` class, or "
                    "an external loop in the `experiment_loop` argument."
                )
            self.loop = options.loops[0]  # TODO: Support nested loops
        else:
            self.loop = experiment_loop
        self.control_bus = control_bus
        self.readout_bus = readout_bus
        super().__init__(platform=platform, circuits=circuits, options=options)

    def post_process_results(self):
        """Method used to post-process the results of an experiment.

        By default this method computes the magnitude of the IQ data and saves it into the ``post_processed_results``
        attribute.

        Returns:
            np.ndarray: post-processed results
        """
        acquisitions = self.results.acquisitions()
        i = np.array(acquisitions["i"])
        q = np.array(acquisitions["q"])
        self.post_processed_results = 20 * np.log10(np.sqrt(i**2 + q**2))
        return self.post_processed_results

    def fit(self, p0: tuple | None = None):
        """Method used to fit the results of an experiment.

        This method uses the scipy function `curve_fit` to fit the function `self.func` to the post-processed data.

        Args:
            p0 (tuple, optional): Initial guess for the parameters. Defaults to None.

        Returns:
            float: Optimal values for the parameters so that the sum of the squared residuals of
                `f(xdata, *popt) - ydata` is minimized.
        """
        if not hasattr(self, "post_processed_results"):
            raise AttributeError(
                "The post-processed results must be computed before fitting. "
                "Please call ``post_process_results`` first."
            )

        self.popt, _ = curve_fit(  # pylint: disable=unbalanced-tuple-unpacking
            self.func, xdata=self.loop.values, ydata=self.post_processed_results, p0=p0
        )

        return self.popt

    def plot(self):
        """Method used to plot the results of an experiment.

        By default this method creates a figure with size (9, 7) and plots the magnitude of the IQ data.
        """
        if not hasattr(self, "post_processed_results"):
            raise AttributeError(
                "The post-processed results must be computed before fitting. "
                "Please call ``post_process_results`` first."
            )
        xdata = self.loop.values

        # Plot data
        fig, axes = plt.subplots(figsize=(9, 7))
        axes.set_title(self.options.name)
        axes.set_xlabel(f"{self.loop.alias}: {self.loop.parameter.value}")
        axes.set_ylabel("|S21| [dB]")  # TODO: Change label for 2D plots
        axes.scatter(xdata, self.post_processed_results, color="blue")
        if hasattr(self, "popt"):
            # Create label text
            args = list(inspect.signature(self.func).parameters.keys())[1:]
            text = "".join(f"{arg}: {value:.5f}\n" for arg, value in zip(args, self.popt))
            axes.plot(xdata, self.func(xdata, *self.popt), "--", label=text, color="red")
            axes.legend(loc="upper right")
        return fig

    def bus_setup(self, parameters: dict[Parameter, float | str | bool], control=False) -> None:
        """Method used to change parameters of the bus used in the experiment. Some possible bus parameters are:

            * Parameter.CURRENT
            * Parameter.ATTENUATION
            * Parameter.IF
            * Parameter.GAIN
            * Parameter.LO_FREQUENCY
            * Parameter.POWER

        Args:
            parameters (dict[Parameter, float | str | bool]): dictionary containing parameter names as keys and
                parameter values as values
            control (bool, optional): whether to change the parameters of the control bus (True) or the readout
                bus (False)
        """
        bus = self.control_bus if control else self.readout_bus

        if bus is None:
            raise ValueError(f"The experiment doesn't have a {'control' if control else 'readout'} bus.")

        for parameter, value in parameters.items():
            bus.set_parameter(parameter=parameter, value=value)

    def gate_setup(self, parameters: dict[Parameter, float | str | bool], gate: str) -> None:
        """Method used to change the parameters of the given gate. Some possible gate parameters are:

            * Parameter.AMPLITUDE
            * Parameter.DURATION
            * Parameter.PHASE

        Args:
            parameters (dict[Parameter, float | str | bool]): dictionary containing parameter names as keys and
                parameter values as values
            gate (str): name of the gate to change
        """
        for parameter, value in parameters.items():
            self.platform.set_parameter(alias=gate, parameter=parameter, value=value)
        self.build_execution()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operation class

This file provides the abstract base class (ABC) Operation, which serves as the base class for all quantum operations in a quantum circuit.

It defines the ParameterValue and Parameters types for handling operation parameters.
The Operation class includes methods for setting and getting parameter values, as well as parsing a string representation of an operation to create an instance of the corresponding operation.
"""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from inspect import Signature, signature

from qililab.circuit.operation_factory import OperationFactory
from qililab.typings import OperationName
from qililab.typings.enums import Qubits
from qililab.utils import classproperty

ParameterValue = int | float | bool
Parameters = dict[str, ParameterValue]


@dataclass(kw_only=True)
class Operation(ABC):
    """Base class of all operations"""

    _class_signature: Signature | None = None

    @classproperty
    @abstractmethod
    def name(self) -> OperationName:
        """Abstract property for the operation name."""

    @classproperty
    @abstractmethod
    def num_qubits(self) -> Qubits:
        """Abstract property for the operation multiplicity."""

    @property
    def parameters(self) -> Parameters:
        """Get the names and values of all parameters as dictionary

        Returns:
            Parameters: The parameters of the operation
        """
        return {}

    @property
    def parameters_names(self) -> tuple[str, ...]:
        """Get the names of all parameters

        Returns:
            tuple[str, ...]: The names of all parameters as tuple
        """
        return tuple(self.parameters.keys())

    @property
    def parameters_values(self) -> tuple[ParameterValue, ...]:
        """Get the values of all parameters

        Returns:
            tuple[ParameterValue, ...]: The values of all parameters as tuple
        """
        return tuple(self.parameters.values())

    def has_parameters(self) -> bool:
        """Returns True if operation has settable parameters, False otherwise

        Returns:
            bool: True if operation has settable parameters, False otherwise
        """
        return len(self.parameters) > 0

    def set_parameter(self, name: str, value: ParameterValue):
        """Set parameter's value

        Args:
            name (str): The name of the parameter
            value (ParameterValue): The new value of the parameter

        Raises:
            ValueError: If parameter does not exist
        """
        if not (hasattr(self, name) and name in self.parameters):
            raise ValueError(f"Operation {self.name} has no parameter '{name}'")
        setattr(self, name, value)

    def get_parameter(self, name: str) -> ParameterValue:
        """Get parameter's value

        Args:
            name (str): The name of the parameter

        Raises:
            ValueError: If parameter does not exist

        Returns:
            ParameterValue: The value of the parameter
        """
        if not (hasattr(self, name) and name in self.parameters):
            raise ValueError(f"Operation {self.name} has no parameter '{name}'")
        return getattr(self, name)

    def __str__(self) -> str:
        """Get the string representation of the operation

        Returns:
            str: The string representation of the operation
        """
        result = self.name.value  # pylint: disable=no-member
        if self.has_parameters():
            parameters = ",".join([f"{name}={value}" for name, value in self.parameters.items()])
            result += f"({parameters})"
        return result

    @classmethod
    def _get_signature(cls) -> Signature:
        """Get the signature of the class. Caches result for quicker retrievals.

        Returns:
            Signature: The signature of the class.
        """
        if cls._class_signature is None:
            cls._class_signature = signature(cls)
        return cls._class_signature

    @classmethod
    def parse(cls, string_representation: str):
        """
        Parse the string representation of an operation and return an instance of the operation.

        Args:
            string_representation (str): String representation of the operation.

        Returns:
            Operation: An instance of the parsed operation.
        """
        if match := re.match(r"(\w+)(?:\((\w+=.+,?)\))?", string_representation):
            operation_name, parameters_str = match.groups()
            operation_class = OperationFactory.get(operation_name)
            operation_signature = operation_class._get_signature()  # pylint: disable=protected-access
            parameters = {}
            if parameters_str is not None:
                for parameter_str in parameters_str.split(","):
                    name, value = parameter_str.split("=")
                    parameter_signature = operation_signature.parameters.get(name)
                    if parameter_signature is None:
                        raise ValueError(f"Operation {operation_name} has no parameter called {name}")
                    value = parameter_signature.annotation(value)
                    parameters[name] = value
            return operation_class(**parameters)
        raise ValueError(f"Invalid string representation: {string_representation}")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Pulsar Controller class"""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentController, InstrumentControllerSettings
from qililab.instruments.qblox.qblox_qcm import QbloxQCM
from qililab.instruments.qblox.qblox_qrm import QbloxQRM
from qililab.typings.enums import InstrumentTypeName, Parameter, ReferenceClock
from qililab.typings.instruments.cluster import Cluster
from qililab.typings.instruments.pulsar import Pulsar


class QbloxController(InstrumentController):
    """Qblox Controller class.

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        settings (QbloxControllerSettings): Settings of the Qblox Pulser Instrument Controller.
    """

    @dataclass
    class QbloxControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific Qblox Pulsar Controller."""

        reference_clock: ReferenceClock

    settings: QbloxControllerSettings
    device: Pulsar | Cluster
    modules: Sequence[QbloxQCM | QbloxQRM]

    @InstrumentController.CheckConnected
    def initial_setup(self):
        """Initial setup"""
        self._set_reference_source()
        super().initial_setup()

    @InstrumentController.CheckConnected
    @abstractmethod
    def _set_reference_source(self):
        """Set reference source. Options are 'internal' or 'external'"""

    @property
    def reference_clock(self):
        """Qblox 'reference_clock' property.

        Returns:
            ReferenceClock: settings.reference_clock.
        """
        return self.settings.reference_clock

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, QbloxQCM) and not isinstance(module, QbloxQRM):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument are {InstrumentTypeName.QBLOX_QCM} "
                    + f"and {InstrumentTypeName.QBLOX_QRM}."
                )

    def to_dict(self):
        """Return a dict representation of the Qblox controller class."""
        return super().to_dict() | {
            Parameter.REFERENCE_CLOCK.value: self.reference_clock.value,
        }



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QiliQasmConvert

This file provides the `QiliQasmConverter` class, which is responsible for converting Qililab quantum circuits
to and from QILIQASM format. The QILIQASM format is a custom representation designed specifically for Qililab circuits.
"""
from qililab.circuit import Circuit  # pylint: disable=cyclic-import
from qililab.circuit.operations import Operation  # pylint: disable=cyclic-import
from qililab.config.version import __version__


class QiliQasmConverter:
    """
    A utility class for converting Qililab quantum circuits to and from QILIQASM format.
    """

    @staticmethod
    def to_qasm(circuit: Circuit):
        """
        Convert a Qililab quantum circuit to QILIQASM format.

        Args:
            circuit (Circuit): The Qililab quantum circuit to convert.

        Returns:
            str: The QILIQASM representation of the given quantum circuit.
        """
        code = [f"// Generated by Qililab {__version__}"]
        code += ["// QILIQASM 1.0"]
        code += [f"Circuit {circuit.num_qubits}"]

        layers = circuit.get_operation_layers()
        for layer in layers:
            for operation_node in layer:
                qubits = ",".join(f"{qubit}" for qubit in operation_node.qubits)
                operation = str(operation_node.operation)
                code += [f"{qubits} {operation}"]

        return "\n".join(code)

    @staticmethod
    def from_qasm(qasm: str):
        """
        Convert a QILIQASM string to a Qililab quantum circuit.

        Args:
            qasm (str): The QILIQASM representation of a quantum circuit.

        Returns:
            Circuit: The Qililab quantum circuit created from the QILIQASM string.
        """
        lines = qasm.split("\n")
        for line in lines:
            if line.startswith("//"):
                continue
            first_str, second_str = line.split(None, 1)
            if first_str == "Circuit":
                num_qubits = int(second_str)
                circuit = Circuit(num_qubits)
            else:
                qubits = tuple(int(qubit) for qubit in first_str.split(","))
                operation = Operation.parse(second_str)
                circuit.add(qubits, operation)
        return circuit



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains all the variables used inside a QProgram."""
import math
from uuid import UUID, uuid4


class Variable:
    """Variable class used to define variables inside a QProgram."""

    _uuid: UUID
    value: int | float

    def __init__(self):
        self._uuid = uuid4()

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)

    def __hash__(self):
        return hash(self._uuid)

    def __format__(self, formatstr):
        return self.value.__format__(formatstr)

    def __pos__(self):
        return +self.value

    def __neg__(self):
        return -self.value

    def __abs__(self):
        return abs(self.value)

    def __round__(self, ndigits=None):
        return round(self.value, ndigits)

    def __floor__(self):
        return math.floor(self.value)

    def __ceil__(self):
        return math.ceil(self.value)

    def __trunc__(self):
        return math.trunc(self.value)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __complex__(self):
        return complex(self.value)

    def __add__(self, other):
        return self.value + other

    def __sub__(self, other):
        return self.value - other

    def __mul__(self, other):
        return self.value * other

    def __truediv__(self, other):
        return self.value / other

    def __floordiv__(self, other):
        return self.value // other

    def __mod__(self, other):
        return self.value % other

    def __pow__(self, other):
        return self.value**other

    def __radd__(self, other):
        return other + self.value

    def __rsub__(self, other):
        return other - self.value

    def __rmul__(self, other):
        return other * self.value

    def __rtruediv__(self, other):
        return other / self.value

    def __rfloordiv__(self, other):
        return other // self.value

    def __rmod__(self, other):
        return other % self.value

    def __rpow__(self, other):
        return other**self.value

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __lt__(self, other):
        return self.value < other

    def __gt__(self, other):
        return self.value > other

    def __le__(self, other):
        return self.value <= other

    def __ge__(self, other):
        return self.value >= other


class IntVariable(Variable, int):  # type: ignore
    """Integer variable. This class is used to define a variable of type int, such that Python recognizes this class
    as an integer."""

    def __init__(self, value: int = 0):
        self.value: int = value
        super().__init__()


class FloatVariable(Variable, float):  # type: ignore
    """Float variable. This class is used to define a variable of type float, such that Python recognizes this class
    as a float."""

    def __init__(self, value: float = 0.0):
        self.value: float = value
        super().__init__()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qibo.gates.abstract import ParametrizedGate


class Wait(ParametrizedGate):
    """The Wait gate.

    Args:
        q (int): the qubit index.
        t (int): the time to wait (ns)
    """

    def __init__(self, q, t):
        super().__init__(trainable=True)
        self.name = "wait"
        self._controlled_gate = None
        self.target_qubits = (q,)

        self.parameters = t
        self.init_args = [q]
        self.init_kwargs = {"t": t}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ClassPropertyDescriptor:  # pylint: disable=too-few-public-methods, disable=missing-class-docstring
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, klass):
        return self.fget.__get__(obj, klass)()


def classproperty(func):  # pylint: disable=missing-function-docstring
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Execute function used to execute a qibo Circuit using the given runcard."""
from qibo.models import Circuit
from tqdm.auto import tqdm

from qililab.result import Result

from .data_management import build_platform
from .transpiler import translate_circuit


def execute(program: Circuit | list[Circuit], runcard: str | dict, nshots: int = 1) -> Result | list[Result]:
    """Executes a Qibo circuit (or a list of circuits) with qililab and returns the results.

    Args:
        circuit (Circuit | list[Circuit]): Qibo Circuit.
        runcard (str | dict): If a string, path to the YAML file containing the serialization of the Platform to be
            used. If a dictionary, the serialized platform to be used.
        nshots (int, optional): Number of shots to execute. Defaults to 1.

    Returns:
        Result | list[Result]: :class:`Result` class (or list of :class:`Result` classes) containing the results of the
            execution.

    Example Usage:

    .. code-block:: python

        from qibo.models import Circuit
        from qibo import gates
        from pathlib import Path
        import qililab as ql
        import os
        import numpy as np

        nqubits = 5
        c = Circuit(nqubits)
        for qubit in range(nqubits):
            c.add(gates.H(qubit))
        c.add(gates.CNOT(2,0))
        c.add(gates.RY(4,np.pi / 4))
        c.add(gates.X(3))
        c.add(gates.M(*range(3)))
        c.add(gates.SWAP(4,2))
        c.add(gates.RX(1, 3*np.pi/2))

        probabilities = ql.execute(c, runcard="./runcards/galadriel.yml")
    """
    if isinstance(program, Circuit):
        program = [program]

    # Initialize platform and connect to the instruments
    platform = build_platform(runcard=runcard)
    platform.connect()

    try:
        platform.initial_setup()
        platform.turn_on_instruments()
        results = []
        for circuit in tqdm(program, total=len(program)):
            # Transpile and optimize circuit
            native_circuit = translate_circuit(circuit, gates_settings=platform.gates_settings)
            # Execute circuit
            results.append(platform.execute(native_circuit, num_avg=1, repetition_duration=200_000, num_bins=nshots))
        platform.disconnect()
        return results[0] if len(results) == 1 else results
    except Exception as e:
        platform.disconnect()
        raise e



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SignalGenerator class."""
from abc import ABC, abstractmethod
from typing import Any


class BaseInstrument(ABC):
    """Base Interface for all instruments."""

    @property
    @abstractmethod
    def params(self):
        """parameters property."""

    @property
    @abstractmethod
    def alias(self):
        """alias property."""

    @abstractmethod
    def set(self, param_name: str, value: Any) -> None:
        """Set instrument parameter.

        Args:
            param_name (str): The name of a parameter of this instrument.
            value (Any): The new value to set.
        """

    @abstractmethod
    def get(self, param_name: str) -> Any:
        """Get instrument parameter.

        Args:
            param_name (str): The name of a parameter of this instrument.

        Returns:
            Any: Current value of the parameter.
        """



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Chip class."""
from dataclasses import asdict, dataclass

import networkx as nx
from networkx import Graph

from qililab.chip.node import Node
from qililab.chip.nodes import Coil, Coupler, Port, Qubit, Resonator
from qililab.constants import RUNCARD
from qililab.typings.enums import Line
from qililab.utils import Factory, dict_factory


@dataclass
class Chip:
    """Chip representation as a graph.
    This class represents the chip structure of in the runcard and contains all the connectivity information of
    the chip.
    """

    nodes: list[Node]

    def __post_init__(self):
        """Cast nodes to their corresponding classes."""
        self.nodes = [Factory.get(name=node.pop(RUNCARD.NAME))(**node) for node in self.nodes]

    def _get_qubit(self, idx: int) -> Qubit:
        """Find qubit from given idx value.

        Args:
            idx (int): Qubit index.

        Raises:
            ValueError: If no qubit is found.

        Returns:
            Qubit: Qubit node object.
        """
        for node in self.nodes:
            if isinstance(node, Qubit) and node.qubit_index == idx:
                return node
        raise ValueError(f"Could not find qubit with idx {idx}.")

    def _get_adjacent_nodes(self, node: Node) -> list[Node | None]:
        """Get adjacent nodes from given node.

        Args:
            node (Node): Node object.

        Returns:
            list[Node | None]: List containing all adjacent nodes.
        """
        return [self.get_node_from_alias(alias=alias) for alias in node.nodes]

    def get_topology(self) -> Graph:
        """Returns a networkx Graph with the qubit connectivity of the chip

        Returns:
            Graph: graph showing the qubit topology
        """
        g = nx.Graph()

        for qubit in self.qubits:
            neighs = [
                neigh
                for neigh in self._get_adjacent_nodes(self.get_node_from_qubit_idx(qubit, readout=False))
                if isinstance(neigh, Qubit)
            ]
            neigh_qubits = [neigh.qubit_index for neigh in neighs if isinstance(neigh, Qubit)]
            edges = [(qubit, neigh_qubit) for neigh_qubit in neigh_qubits]
            g.add_edges_from(edges)
        return g

    def get_node_from_qubit_idx(self, idx: int, readout: bool) -> Qubit | Resonator:
        """Get node class from qubit index.

        Args:
            idx (int): Qubit index.
            readout (bool): If True, return readout port and resonator frequency, if False return control port and qubit
                frequency.

        Raises:
            ValueError: if qubit doesn't have a readout line

        Returns:
            Qubit | Resonator: qubit/resonator with the given qubit index
        """
        qubit = self._get_qubit(idx=idx)
        if not readout:
            return qubit
        adj_nodes = self._get_adjacent_nodes(node=qubit)
        for node in adj_nodes:
            if isinstance(node, Resonator):
                return node
        raise ValueError(f"Qubit with index {idx} doesn't have a readout line.")

    def get_port_from_qubit_idx(self, idx: int, line: Line) -> str:
        """Find Qubit's port for specific line type

        Args:
            idx (int): Qubit index.
            line (Line): The type of line

        Raises:
            ValueError: If qubit isn't connected to this type of line

        Returns:
            str: The alias of the port
        """
        readout = line in [Line.FEEDLINE_INPUT, Line.FEEDLINE_OUTPUT]
        node = self.get_node_from_qubit_idx(idx=idx, readout=readout)
        adjacent_nodes = self._get_adjacent_nodes(node=node)

        for adjacent_node in adjacent_nodes:
            if isinstance(adjacent_node, Port) and adjacent_node.line == line:
                return adjacent_node.alias

        raise ValueError(f"Qubit with index {idx} doesn't have a {line} line.")

    def get_port_nodes(self, alias: str) -> list[Qubit | Resonator | Coupler | Coil]:
        """Get nodes connected to a given port.

        Args:
            alias (str): Alias of the port.

        Returns:
            list[Node]: List of nodes connected to the given port.
        """
        port = self.get_node_from_alias(alias=alias)
        return self._get_adjacent_nodes(node=port)  # type: ignore

    def get_node_from_alias(self, alias: str) -> Node | None:
        """Get node from given alias.

        Args:
            alias (str): Alias of the node.

        Raises:
            ValueError: If no node is found.

        Returns:
            Node | None: Node class.
        """
        for node in self.nodes:
            if node.alias == alias:
                return node
        return None

    def to_dict(self):
        """Return a dict representation of the Chip class."""
        return {
            "nodes": [{RUNCARD.NAME: node.name.value} | asdict(node, dict_factory=dict_factory) for node in self.nodes]
        }

    @property
    def qubits(self):
        """Chip `qubits` property.

        Returns:
            list[int]: List of integers containing the indices of the qubits inside the chip.
        """
        return [node.qubit_index for node in self.nodes if isinstance(node, Qubit)]

    @property
    def num_qubits(self) -> int:
        """Chip 'num_qubits' property

        Returns:
            int: Number of qubits.
        """
        return sum(isinstance(node, Qubit) for node in self.nodes)

    @property
    def num_ports(self) -> int:
        """Chip 'num_ports' property

        Returns:
            int: Number of ports.
        """
        return sum(isinstance(node, Port) for node in self.nodes)

    def __str__(self):
        """String representation of the Chip class."""
        string = f"Chip with {self.num_qubits} qubits and {self.num_ports} ports: \n\n"
        for node in self.nodes:
            if isinstance(node, Port):
                adj_nodes = self._get_adjacent_nodes(node=node)
                string += f" * Port {node.alias} ({node.line.value}): ----"
                for adj_node in adj_nodes:
                    string += f"|{adj_node}|--"
                string += "--\n"
        return string



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Typings from AWG Types """

from enum import Enum


class AWGTypes(Enum):
    """Typings from AWG Types"""

    AWG_SEQUENCERS = "awg_sequencers"
    OUT_OFFSETS = "out_offsets"


class AWGSequencerTypes(Enum):
    """Types from AWG Sequencer Types"""

    IDENTIFIER = "identifier"
    INTERMEDIATE_FREQUENCY = "intermediate_frequency"
    OFFSET_I = "offset_i"
    OFFSET_Q = "offset_q"
    CHIP_PORT_ID = "chip_port_id"


class AWGIQChannelTypes(Enum):
    """Types from AWG IQ Channel Types"""

    IDENTIFIER = "identifier"
    I_CHANNEL = "i_channel"
    Q_CHANNEL = "q_channel"


class AWGChannelMappingTypes(Enum):
    """Types from AWG Channel Mapping Types"""

    AWG_SEQUENCER_IDENTIFIER = "awg_sequencer_identifier"
    AWG_SEQUENCER_PATH_IDENTIFIER = "awg_sequencer_path_identifier"


class AWGOutputChannelTypes(Enum):
    """Types from AWG Ouput Channel Types"""

    IDENTIFIER = "identifier"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox module class"""
import itertools
from abc import abstractmethod
from dataclasses import dataclass
from typing import Sequence, cast

import numpy as np
from qpysequence import Acquisitions, Program
from qpysequence import Sequence as QpySequence
from qpysequence import Waveforms, Weights
from qpysequence.library import long_wait
from qpysequence.program import Block, Loop, Register
from qpysequence.program.instructions import Play, ResetPh, SetAwgGain, SetPh, Stop
from qpysequence.utils.constants import AWG_MAX_GAIN

from qililab.config import logger
from qililab.instruments.awg import AWG
from qililab.instruments.awg_settings import AWGQbloxSequencer
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.pulse import PulseBusSchedule, PulseShape
from qililab.typings.enums import Parameter
from qililab.typings.instruments import Pulsar, QcmQrm


class QbloxModule(AWG):
    """Qblox Module class.

    Args:
        device (Pulsar): Instance of the Qblox Pulsar class used to connect to the instrument.
        settings (QbloxPulsarSettings): Settings of the instrument.
    """

    _MAX_BINS: int = 131072
    _NUM_MAX_SEQUENCERS: int = 6
    _NUM_MAX_AWG_OUT_CHANNELS: int = 4
    _MIN_WAIT_TIME: int = 4  # in ns

    @dataclass
    class QbloxModuleSettings(AWG.AWGSettings):
        """Contains the settings of a specific pulsar.

        Args:
            awg_sequencers (Sequence[AWGQbloxSequencer]): list of settings for each sequencer
            out_offsets (list[float]): list of offsets for each output of the qblox module
        """

        awg_sequencers: Sequence[AWGQbloxSequencer]
        out_offsets: list[float]

        def __post_init__(self):
            """build AWGQbloxSequencer"""
            if (
                self.num_sequencers <= 0
                or self.num_sequencers > QbloxModule._NUM_MAX_SEQUENCERS  # pylint: disable=protected-access
            ):
                raise ValueError(
                    "The number of sequencers must be greater than 0 and less or equal than "
                    + f"{QbloxModule._NUM_MAX_SEQUENCERS}. Received: {self.num_sequencers}"  # pylint: disable=protected-access
                )
            if len(self.awg_sequencers) != self.num_sequencers:
                raise ValueError(
                    f"The number of sequencers: {self.num_sequencers} does not match"
                    + f" the number of AWG Sequencers settings specified: {len(self.awg_sequencers)}"
                )

            self.awg_sequencers = [
                AWGQbloxSequencer(**sequencer)
                if isinstance(sequencer, dict)
                else sequencer  # pylint: disable=not-a-mapping
                for sequencer in self.awg_sequencers
            ]
            super().__post_init__()

    settings: QbloxModuleSettings
    device: Pulsar | QcmQrm
    # Cache containing the last compiled pulse schedule for each sequencer
    _cache: dict[int, PulseBusSchedule] = {}

    def __init__(self, settings: dict):
        # The sequences dictionary contains all the compiled sequences for each sequencer and a flag indicating whether
        # the sequence has been uploaded or not
        self.sequences: dict[int, tuple[QpySequence, bool]] = {}  # {sequencer_idx: (program, True), ...}
        # TODO: Set this attribute during initialization of the instrument
        self.nshots: int | None = None
        self.num_bins: int = 1
        self.repetition_duration: int | None = None
        super().__init__(settings=settings)

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """Initial setup"""
        self._map_outputs()
        self.clear_cache()
        for sequencer in self.awg_sequencers:
            sequencer_id = sequencer.identifier
            # Set `sync_en` flag to False (this value will be set to True if the sequencer is used in the execution)
            self.device.sequencers[sequencer_id].sync_en(False)
            self._set_nco(sequencer_id=sequencer_id)
            self._set_gain_i(value=sequencer.gain_i, sequencer_id=sequencer_id)
            self._set_gain_q(value=sequencer.gain_q, sequencer_id=sequencer_id)
            self._set_offset_i(value=sequencer.offset_i, sequencer_id=sequencer_id)
            self._set_offset_q(value=sequencer.offset_q, sequencer_id=sequencer_id)
            self._set_hardware_modulation(value=sequencer.hardware_modulation, sequencer_id=sequencer_id)
            self._set_gain_imbalance(value=sequencer.gain_imbalance, sequencer_id=sequencer_id)
            self._set_phase_imbalance(value=sequencer.phase_imbalance, sequencer_id=sequencer_id)
            ALL_ON = 15  # 1111 in binary
            self._set_markers(value=ALL_ON, sequencer_id=sequencer_id)

        for idx, offset in enumerate(self.out_offsets):
            self._set_out_offset(output=idx, value=offset)

    def desync_sequencers(self) -> None:
        """Desyncs all sequencers."""
        for sequencer in self.awg_sequencers:
            self.device.sequencers[sequencer.identifier].sync_en(False)

    @property
    def module_type(self):
        """returns the qblox module type. Options: QCM or QRM"""
        return self.device.module_type()

    def compile(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ) -> list[QpySequence]:
        """Compiles the ``PulseBusSchedule`` into an assembly program.

        This method skips compilation if the pulse schedule is in the cache. Otherwise, the pulse schedule is
        compiled and added into the cache.

        If the number of shots or the repetition duration changes, the cache will be cleared.

        Args:
            pulse_bus_schedule (PulseBusSchedule): the list of pulses to be converted into a program
            nshots (int): number of shots / hardware average
            repetition_duration (int): repetition duration
            num_bins (int): number of bins

        Returns:
            list[QpySequence]: list of compiled assembly programs
        """
        if nshots != self.nshots or repetition_duration != self.repetition_duration or num_bins != self.num_bins:
            self.nshots = nshots
            self.repetition_duration = repetition_duration
            self.num_bins = num_bins
            self.clear_cache()

        compiled_sequences = []
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=pulse_bus_schedule.port)
        for sequencer in sequencers:
            if pulse_bus_schedule != self._cache.get(sequencer.identifier):
                sequence = self._compile(pulse_bus_schedule, sequencer)
                compiled_sequences.append(sequence)
            else:
                compiled_sequences.append(self.sequences[sequencer.identifier][0])
        return compiled_sequences

    def _compile(self, pulse_bus_schedule: PulseBusSchedule, sequencer: AWGQbloxSequencer) -> QpySequence:
        """Compiles the ``PulseBusSchedule`` into an assembly program and updates the cache and the saved sequences.

        Args:
            pulse_bus_schedule (PulseBusSchedule): the list of pulses to be converted into a program
            sequencer (int): index of the sequencer to generate the program
        """
        sequence = self._translate_pulse_bus_schedule(pulse_bus_schedule=pulse_bus_schedule, sequencer=sequencer)
        self._cache[sequencer.identifier] = pulse_bus_schedule
        self.sequences[sequencer.identifier] = (sequence, False)
        return sequence

    def run(self, port: str):
        """Run the uploaded program"""
        self.start_sequencer(port=port)

    def _translate_pulse_bus_schedule(self, pulse_bus_schedule: PulseBusSchedule, sequencer: AWGQbloxSequencer):
        """Translate a pulse sequence into a Q1ASM program and a waveform dictionary.

        Args:
            pulse_bus_schedule (PulseBusSchedule): Pulse bus schedule to translate.
            sequencer (int): index of the sequencer to generate the program

        Returns:
            Sequence: Qblox Sequence object containing the program and waveforms.
        """
        waveforms = self._generate_waveforms(pulse_bus_schedule=pulse_bus_schedule, sequencer=sequencer)
        acquisitions = self._generate_acquisitions()
        program = self._generate_program(
            pulse_bus_schedule=pulse_bus_schedule, waveforms=waveforms, sequencer=sequencer.identifier
        )
        weights = self._generate_weights(sequencer=sequencer)
        return QpySequence(program=program, waveforms=waveforms, acquisitions=acquisitions, weights=weights)

    def _generate_program(  # pylint: disable=too-many-locals
        self, pulse_bus_schedule: PulseBusSchedule, waveforms: Waveforms, sequencer: int
    ):
        """Generate Q1ASM program

        Args:
            pulse_sequence (PulseSequence): pulse sequence
            waveforms (Waveforms): waveforms
            sequencer (int): index of the sequencer to generate the program

        Returns:
            Program: Q1ASM program.
        """

        # Define program's blocks
        program = Program()
        start = Block(name="start")
        start.append_component(ResetPh())
        program.append_block(block=start)
        # Create registers with 0 and 1 (necessary for qblox)
        weight_registers = Register(), Register()
        self._init_weights_registers(registers=weight_registers, values=(0, 1), program=program)
        avg_loop = Loop(name="average", begin=int(self.nshots))  # type: ignore
        bin_loop = Loop(name="bin", begin=0, end=self.num_bins, step=1)
        avg_loop.append_component(bin_loop)
        program.append_block(avg_loop)
        stop = Block(name="stop")
        stop.append_component(Stop())
        program.append_block(block=stop)
        timeline = pulse_bus_schedule.timeline
        if len(timeline) > 0 and timeline[0].start_time != 0:
            bin_loop.append_component(long_wait(wait_time=int(timeline[0].start_time)))

        for i, pulse_event in enumerate(timeline):
            waveform_pair = waveforms.find_pair_by_name(pulse_event.pulse.label())
            wait_time = timeline[i + 1].start_time - pulse_event.start_time if (i < (len(timeline) - 1)) else 4
            phase = int((pulse_event.pulse.phase % (2 * np.pi)) * 1e9 / (2 * np.pi))
            gain = int(np.abs(pulse_event.pulse.amplitude) * AWG_MAX_GAIN)  # np.abs() needed for negative pulses
            bin_loop.append_component(SetAwgGain(gain_0=gain, gain_1=gain))
            bin_loop.append_component(SetPh(phase=phase))
            bin_loop.append_component(
                Play(
                    waveform_0=waveform_pair.waveform_i.index,
                    waveform_1=waveform_pair.waveform_q.index,
                    wait_time=int(wait_time),
                )
            )
        self._append_acquire_instruction(
            loop=bin_loop, bin_index=bin_loop.counter_register, sequencer_id=sequencer, weight_regs=weight_registers
        )
        if self.repetition_duration is not None:
            wait_time = self.repetition_duration - bin_loop.duration_iter
            if wait_time > self._MIN_WAIT_TIME:
                bin_loop.append_component(long_wait(wait_time=wait_time))

        logger.info("Q1ASM program: \n %s", repr(program))  # pylint: disable=protected-access
        return program

    def _init_weights_registers(self, registers: tuple[Register, Register], values: tuple[int, int], program: Program):
        """Initialize the weights `registers` to the `values` specified and place the required instructions in the
        setup block of the `program`."""

    def _generate_acquisitions(self) -> Acquisitions:
        """Generate Acquisitions object, currently containing a single acquisition named "default", with num_bins = 1
        and index = 0.

        Returns:
            Acquisitions: Acquisitions object.
        """
        # FIXME: is it really necessary to generate acquisitions for a QCM??
        acquisitions = Acquisitions()
        acquisitions.add(name="default", num_bins=self.num_bins, index=0)
        return acquisitions

    @abstractmethod
    def _generate_weights(self, sequencer: AWGQbloxSequencer) -> Weights:
        """Generate acquisition weights.

        Returns:
            dict: Acquisition weights.
        """

    @abstractmethod
    def _append_acquire_instruction(
        self, loop: Loop, bin_index: Register | int, sequencer_id: int, weight_regs: tuple[Register, Register]
    ):
        """Append an acquire instruction to the loop."""

    def start_sequencer(self, port: str):
        """Start sequencer and execute the uploaded instructions."""
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            if sequencer.identifier in self.sequences:
                self.device.arm_sequencer(sequencer=sequencer.identifier)
                self.device.start_sequencer(sequencer=sequencer.identifier)

    @Instrument.CheckDeviceInitialized
    def setup(  # pylint: disable=too-many-branches, too-many-return-statements
        self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None, port_id: str | None = None
    ):
        """Set Qblox instrument calibration settings."""
        if parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            self._set_out_offset(output=output, value=value)
            return

        if channel_id is None:
            if port_id is not None:
                channel_id = self.get_sequencers_from_chip_port_id(chip_port_id=port_id)[0].identifier
            elif self.num_sequencers == 1:
                channel_id = 0
            else:
                raise ParameterNotFound(f"Cannot update parameter {parameter.value} without specifying a channel_id.")

        if channel_id > self.num_sequencers - 1:
            raise ParameterNotFound(
                f"the specified channel id:{channel_id} is out of range. Number of sequencers is {self.num_sequencers}"
            )
        if parameter == Parameter.GAIN:
            self._set_gain(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_I:
            self._set_gain_i(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_Q:
            self._set_gain_q(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.OFFSET_I:
            self._set_offset_i(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.OFFSET_Q:
            self._set_offset_q(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.IF:
            self._set_frequency(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.HARDWARE_MODULATION:
            self._set_hardware_modulation(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.NUM_BINS:
            self._set_num_bins(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.GAIN_IMBALANCE:
            self._set_gain_imbalance(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.PHASE_IMBALANCE:
            self._set_phase_imbalance(value=value, sequencer_id=channel_id)
            return
        raise ParameterNotFound(f"Invalid Parameter: {parameter.value}")

    def get(self, parameter: Parameter, channel_id: int | None = None, port_id: str | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            return self.out_offsets[output]

        if channel_id is None:
            if port_id is not None:
                channel_id = self.get_sequencers_from_chip_port_id(chip_port_id=port_id)[0].identifier
            elif self.num_sequencers == 1:
                channel_id = 0
            else:
                raise ParameterNotFound(f"Cannot update parameter {parameter.value} without specifying a channel_id.")

        sequencer = self._get_sequencer_by_id(id=channel_id)

        if parameter == Parameter.GAIN:
            return sequencer.gain_i, sequencer.gain_q

        if hasattr(sequencer, parameter.value):
            return getattr(sequencer, parameter.value)

        raise ParameterNotFound(f"Cannot find parameter {parameter.value} in instrument {self.alias}")

    @Instrument.CheckParameterValueFloatOrInt
    def _set_num_bins(self, value: float | str | bool, sequencer_id: int):
        """set num_bins for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        if int(value) > self._MAX_BINS:
            raise ValueError(f"Value {value} greater than maximum bins: {self._MAX_BINS}")
        cast(AWGQbloxSequencer, self._get_sequencer_by_id(id=sequencer_id)).num_bins = int(value)

    @Instrument.CheckParameterValueBool
    def _set_hardware_modulation(self, value: float | str | bool, sequencer_id: int):
        """set hardware modulation

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        self._get_sequencer_by_id(id=sequencer_id).hardware_modulation = bool(value)
        self.device.sequencers[sequencer_id].mod_en_awg(bool(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_frequency(self, value: float | str | bool, sequencer_id: int):
        """set frequency

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._get_sequencer_by_id(id=sequencer_id).intermediate_frequency = float(value)
        self.device.sequencers[sequencer_id].nco_freq(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_offset_i(self, value: float | str | bool, sequencer_id: int):
        """Set the offset of the I channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).offset_i = float(value)
        # update value in the instrument
        path = self._get_sequencer_by_id(id=sequencer_id).path_i
        sequencer = self.device.sequencers[sequencer_id]
        getattr(sequencer, f"offset_awg_path{path}")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_offset_q(self, value: float | str | bool, sequencer_id: int):
        """Set the offset of the Q channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).offset_q = float(value)
        # update value in the instrument
        path = self._get_sequencer_by_id(id=sequencer_id).path_q
        sequencer = self.device.sequencers[sequencer_id]
        getattr(sequencer, f"offset_awg_path{path}")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_out_offset(self, output: int, value: float | str | bool):
        """Set output offsets of the Qblox device.

        Args:
            output (int): output to update
            value (float | str | bool): value to update

        Raises:
            ValueError: when value type is not float or int
        """
        if output > len(self.out_offsets):
            raise IndexError(
                f"Output {output} is out of range. The runcard has only {len(self.out_offsets)} output offsets defined."
                " Please update the list of output offsets of the runcard such that it contains a value for each "
                "output of the device."
            )
        self.out_offsets[output] = value
        getattr(self.device, f"out{output}_offset")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain_i(self, value: float | str | bool, sequencer_id: int):
        """Set the gain of the I channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).gain_i = float(value)
        # update value in the instrument
        path = self._get_sequencer_by_id(id=sequencer_id).path_i
        sequencer = self.device.sequencers[sequencer_id]
        getattr(sequencer, f"gain_awg_path{path}")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain_q(self, value: float | str | bool, sequencer_id: int):
        """Set the gain of the Q channel of the given sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        # update value in qililab
        self._get_sequencer_by_id(id=sequencer_id).gain_q = float(value)
        # update value in the instrument
        path = self._get_sequencer_by_id(id=sequencer_id).path_q
        sequencer = self.device.sequencers[sequencer_id]
        getattr(sequencer, f"gain_awg_path{path}")(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain(self, value: float | str | bool, sequencer_id: int):
        """set gain

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._set_gain_i(value=value, sequencer_id=sequencer_id)
        self._set_gain_q(value=value, sequencer_id=sequencer_id)

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Stop the QBlox sequencer from sending pulses."""
        for seq_idx in range(self.num_sequencers):
            self.device.stop_sequencer(sequencer=seq_idx)

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Turn on an instrument."""

    def clear_cache(self):
        """Empty cache."""
        self._cache = {}
        self.sequences = {}

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Reset instrument."""
        self.clear_cache()
        self.device.reset()

    def upload(self, port: str):
        """Upload all the previously compiled programs to its corresponding sequencers.

        This method must be called after the method ``compile``."""
        if self.nshots is None or self.repetition_duration is None:
            raise ValueError("Please compile the circuit before uploading it to the device.")
        sequencers = self.get_sequencers_from_chip_port_id(chip_port_id=port)
        for sequencer in sequencers:
            if (seq_idx := sequencer.identifier) in self.sequences:
                sequence, uploaded = self.sequences[seq_idx]
                self.device.sequencers[seq_idx].sync_en(True)
                if not uploaded:
                    logger.info("Sequence program: \n %s", repr(sequence._program))  # pylint: disable=protected-access
                    self.device.sequencers[seq_idx].sequence(sequence.todict())
                    self.sequences[seq_idx] = (sequence, True)

    def _set_nco(self, sequencer_id: int):
        """Enable modulation of pulses and setup NCO frequency."""
        if self._get_sequencer_by_id(id=sequencer_id).hardware_modulation:
            self._set_hardware_modulation(
                value=self._get_sequencer_by_id(id=sequencer_id).hardware_modulation, sequencer_id=sequencer_id
            )
            self._set_frequency(
                value=self._get_sequencer_by_id(id=sequencer_id).intermediate_frequency, sequencer_id=sequencer_id
            )

    @Instrument.CheckParameterValueFloatOrInt
    def _set_gain_imbalance(self, value: float | str | bool, sequencer_id: int):
        """Set I and Q gain imbalance of sequencer.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """

        self._get_sequencer_by_id(id=sequencer_id).gain_imbalance = float(value)
        self.device.sequencers[sequencer_id].mixer_corr_gain_ratio(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_phase_imbalance(self, value: float | str | bool, sequencer_id: int):
        """Set I and Q phase imbalance of sequencer.

         Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self._get_sequencer_by_id(id=sequencer_id).phase_imbalance = float(value)
        self.device.sequencers[sequencer_id].mixer_corr_phase_offset_degree(float(value))

    @Instrument.CheckParameterValueFloatOrInt
    def _set_markers(self, value: int, sequencer_id: int):
        """Set markers ON/OFF on qblox modules.

        For the RF modules, this command is also used to enable/disable:
            - The 2 outputs (for the QCM-RF).
            - The input and the output (for QRM-RF).

         Args:
            value (int): ON/OFF of the 4 markers in binary (range: 0-15 -> (0000)-(1111)). For the RF modules, the
                first 2 bits correspond to the ON/OFF value of the outputs/inputs and the last 2 bits correspond
                to the 2 markers.
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not int
        """
        self.device.sequencers[sequencer_id].marker_ovr_en(True)
        self.device.sequencers[sequencer_id].marker_ovr_value(value)

    def _map_outputs(self):
        """Disable all connections and map sequencer paths with output channels."""
        # Disable all connections
        for sequencer, out in itertools.product(self.device.sequencers, range(self._NUM_MAX_SEQUENCERS)):
            if hasattr(sequencer, f"channel_map_path{out % 2}_out{out}_en"):
                sequencer.set(f"channel_map_path{out % 2}_out{out}_en", False)

        for sequencer in self.awg_sequencers:
            if sequencer.output_i is not None:
                self.device.sequencers[sequencer.identifier].set(
                    f"channel_map_path{sequencer.path_i}_out{sequencer.output_i}_en", True
                )
            if sequencer.output_q is not None:
                self.device.sequencers[sequencer.identifier].set(
                    f"channel_map_path{sequencer.path_q}_out{sequencer.output_q}_en", True
                )

    def _generate_waveforms(self, pulse_bus_schedule: PulseBusSchedule, sequencer: AWGQbloxSequencer):
        """Generate I and Q waveforms from a PulseSequence object.
        Args:
            pulse_bus_schedule (PulseBusSchedule): PulseSequence object.
        Returns:
            Waveforms: Waveforms object containing the generated waveforms.
        """
        waveforms = Waveforms()

        unique_pulses: list[tuple[int, PulseShape]] = []

        for pulse_event in pulse_bus_schedule.timeline:
            if (pulse_event.duration, pulse_event.pulse.pulse_shape) not in unique_pulses:
                unique_pulses.append((pulse_event.duration, pulse_event.pulse.pulse_shape))
                amp = pulse_event.pulse.amplitude
                sign = 1 if amp >= 0 else -1
                envelope = pulse_event.envelope(amplitude=sign * 1.0)
                real = np.real(envelope)
                imag = np.imag(envelope)
                pair = (real, imag)
                if (sequencer.path_i, sequencer.path_q) == (1, 0):
                    pair = pair[::-1]  # swap paths
                waveforms.add_pair(pair=pair, name=pulse_event.pulse.label())

        return waveforms

    @property
    def final_wait_time(self) -> int:
        """QbloxPulsar 'final_wait_time' property.

        Returns:
            int: Final wait time.
        """
        return self._MIN_WAIT_TIME

    @property
    def out_offsets(self):
        """Returns the offsets of each output of the qblox module."""
        return self.settings.out_offsets

    def _get_sequencer_by_id(self, id: int):  # pylint: disable=redefined-builtin
        """Returns a sequencer with the given `id`."

        Args:
            id (int): Id of the sequencer.

        Raises:
            IndexError: There is no sequencer with the given `id`.

        Returns:
            AWGQbloxSequencer: Sequencer with the given `id`.
        """
        for sequencer in self.awg_sequencers:
            if sequencer.identifier == id:
                return sequencer
        raise IndexError(f"There is no sequencer with id={id}.")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains our native Park gate."""
from qibo.gates import Gate


class Park(Gate):
    """Parking gate qibo interface.
    Args:
        q (int): the qubit id number.
    """

    def __init__(self, q):
        super().__init__()
        self.name = "park"
        self.target_qubits = (q,)
        self.init_args = [q]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Runcard class."""
import ast
import re
from dataclasses import asdict, dataclass
from typing import Literal

from qililab.constants import GATE_ALIAS_REGEX
from qililab.settings.gate_event_settings import GateEventSettings
from qililab.typings.enums import OperationTimingsCalculationMethod, Parameter, ResetMethod
from qililab.utils import nested_dataclass

from .settings import Settings

# pylint: disable=too-few-public-methods


@nested_dataclass
class Runcard:
    """Runcard class. Casts the platform dictionary into a class.

    The input to the constructor should be a dictionary of the desired runcard with the following structure:
    - gates_settings:
    - chip:
    - buses:
    - instruments: List of "instruments" dictionaries
    - instrument_controllers: List of "instrument_controllers" dictionaries

    The gates_settings, chip and bus dictionaries will be passed to their corresponding Runcard.GatesSettings,
    Runcard.Chip or Runcard.Bus classes here, meanwhile the instruments and instrument_controllers will remain dictionaries.

    Then this full class gets passed to the Platform who will instantiate the actual qililab Chip, Buses/Bus and the
    corresponding Instrument classes with the settings attributes of this class.

    Args:
        gates_settings (dict): Gates settings dictionary -> Runcard.GatesSettings inner dataclass
        chip (dict): Chip settings dictionary -> Runcard.Chip settings inner dataclass
        buses (list[dict]): List of Bus settings dictionaries -> list[Runcard.Bus] settings inner dataclass
        instruments (list[dict]): List of dictionaries containing the "instruments" information (does not transform)
        instruments_controllers (list[dict]): List of dictionaries containing the "instrument_controllers" information
            (does not transform)
    """

    # Inner dataclasses definition
    @dataclass
    class Bus:
        """Dataclass with all the settings the buses of the platform need.

        Args:
            alias (str): Alias of the bus.
            system_control (dict): Dictionary containing the settings of the system control of the bus.
            port (str): Alias of the port of the chip the bus is connected to.
            distortions (list[dict]): List of dictionaries containing the settings of the distortions applied to each
                bus.
            delay (int, optional): Delay applied to all pulses sent in this bus. Defaults to 0.
        """

        alias: str
        system_control: dict
        port: str
        distortions: list[dict]
        delay: int = 0

    @dataclass
    class Chip:
        """Dataclass with all the settings/nodes the chip of the platform needs.

        Args:
            nodes (list[dict]): List of dictionaries containing the settings of all the nodes of the chip.
        """

        nodes: list[dict]

    @nested_dataclass
    class GatesSettings(Settings):
        """Dataclass with all the settings and gates definitions needed to decompose gates into pulses."""

        @nested_dataclass
        class OperationSettings:
            """Dataclass with all the settings an operation needs."""

            @dataclass
            class PulseSettings:
                """Dataclass with all the settings a pulse needs."""

                name: str
                amplitude: float
                duration: int
                parameters: dict

            name: str
            pulse: PulseSettings

        minimum_clock_time: int
        delay_between_pulses: int
        delay_before_readout: int
        timings_calculation_method: Literal[
            OperationTimingsCalculationMethod.AS_SOON_AS_POSSIBLE, OperationTimingsCalculationMethod.AS_LATE_AS_POSSIBLE
        ]
        reset_method: Literal[ResetMethod.ACTIVE, ResetMethod.PASSIVE]
        passive_reset_duration: int
        operations: list[OperationSettings]
        gates: dict[str, list[GateEventSettings]]

        def __post_init__(self):
            """Build the Gates Settings based on the master settings."""
            self.gates = {
                gate: [GateEventSettings(**event) for event in schedule] for gate, schedule in self.gates.items()
            }

        def to_dict(self):
            """Serializes gate settings to dictionary and removes fields with None values"""

            def remove_none_values(data):
                if isinstance(data, dict):
                    data = {key: remove_none_values(item) for key, item in data.items() if item is not None}
                elif isinstance(data, list):
                    data = [remove_none_values(item) for item in data if item is not None]
                return data

            return remove_none_values(data=asdict(self))

        def get_operation_settings(self, name: str) -> OperationSettings:
            """Get OperationSettings by operation's name.

            Args:
                name (str): Name of the operation

            Raises:
                ValueError: If no operation is found

            Returns:
                OperationSettings: Operation's settings
            """
            for operation in self.operations:
                # TODO: Fix bug that parses settings as dict instead of defined classes
                if isinstance(operation, dict):
                    operation = Runcard.GatesSettings.OperationSettings(**operation)
                if operation.name == name:
                    return operation
            raise ValueError(f"Operation {name} not found in gates settings.")

        def get_gate(self, name: str, qubits: int | tuple[int, int] | tuple[int]):
            """Get gates settings from runcard for a given gate name and qubits.

            Args:
                name (str): Name of the gate.
                qubits (int |  tuple[int, int] | tuple[int]): The qubits the gate is acting on.

            Raises:
                ValueError: If no gate is found.

            Returns:
                GatesSettings: gate settings.
            """

            gate_qubits = (
                (qubits,) if isinstance(qubits, int) else qubits
            )  # tuplify so that the join method below is general
            gate_name = f"{name}({', '.join(map(str, gate_qubits))})"
            gate_name_t = f"{name}({', '.join(map(str, gate_qubits[::-1]))})"

            # parse spaces in tuple if needed, check first case with spaces since it is more common
            if gate_name.replace(" ", "") in self.gates.keys():
                return self.gates[gate_name.replace(" ", "")]
            if gate_name in self.gates.keys():
                return self.gates[gate_name]
            if gate_name_t.replace(" ", "") in self.gates.keys():
                return self.gates[gate_name_t.replace(" ", "")]
            if gate_name_t in self.gates.keys():
                return self.gates[gate_name_t]
            raise KeyError(f"Gate {name} for qubits {qubits} not found in settings.")

        @property
        def gate_names(self) -> list[str]:
            """GatesSettings 'gate_names' property.

            Returns:
                list[str]: List of the names of all the defined gates.
            """
            return list(self.gates.keys())

        def set_parameter(
            self,
            parameter: Parameter,
            value: float | str | bool,
            channel_id: int | None = None,
            alias: str | None = None,
        ):
            """Cast the new value to its corresponding type and set the new attribute."""
            if alias is None or alias == "platform":
                super().set_parameter(parameter=parameter, value=value, channel_id=channel_id)
                return
            regex_match = re.search(GATE_ALIAS_REGEX, alias)
            if regex_match is None:
                raise ValueError(f"Alias {alias} has incorrect format")
            name = regex_match["gate"]
            qubits_str = regex_match["qubits"]
            qubits = ast.literal_eval(qubits_str)
            gates_settings = self.get_gate(name=name, qubits=qubits)
            schedule_element = 0 if len(alias.split("_")) == 1 else int(alias.split("_")[1])
            gates_settings[schedule_element].set_parameter(parameter, value)

        def get_parameter(
            self,
            parameter: Parameter,
            channel_id: int | None = None,
            alias: str | None = None,
        ):
            """Get parameter from gate settings.

            Args:
                parameter (Parameter): Name of the parameter to get.
                channel_id (int | None, optional): Channel id. Defaults to None.
                alias (str): String which specifies where the parameter can be found.
            """
            if alias is None or alias == "platform":
                return super().get_parameter(parameter=parameter, channel_id=channel_id)
            regex_match = re.search(GATE_ALIAS_REGEX, alias)
            if regex_match is None:
                raise ValueError(f"Could not find gate {alias} in gate settings.")
            name = regex_match["gate"]
            qubits_str = regex_match["qubits"]
            qubits = ast.literal_eval(qubits_str)
            gates_settings = self.get_gate(name=name, qubits=qubits)
            schedule_element = 0 if len(alias.split("_")) == 1 else int(alias.split("_")[1])
            return gates_settings[schedule_element].get_parameter(parameter)

    # Runcard class actual initialization
    name: str
    device_id: int
    chip: Chip
    buses: list[Bus]  # This actually is a list[dict] until the post_init is called
    instruments: list[dict]
    instrument_controllers: list[dict]
    gates_settings: GatesSettings

    def __post_init__(self):
        self.buses = [self.Bus(**bus) for bus in self.buses] if self.buses is not None else None



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the QbloxQCMRF class."""
from dataclasses import dataclass, field

from qililab.instruments.awg_settings import AWGQbloxSequencer  # pylint: disable=cyclic-import
from qililab.instruments.instrument import Instrument, ParameterNotFound  # pylint: disable=cyclic-import
from qililab.instruments.utils.instrument_factory import InstrumentFactory  # pylint: disable=cyclic-import
from qililab.typings import InstrumentName, Parameter

from .qblox_qcm import QbloxQCM


@InstrumentFactory.register
class QbloxQCMRF(QbloxQCM):
    """Qblox QCM-RF driver."""

    name = InstrumentName.QCMRF

    @dataclass
    class QbloxQCMRFSettings(QbloxQCM.QbloxQCMSettings):  # pylint: disable=too-many-instance-attributes
        """Contains the settings of a specific Qblox QCM-RF module."""

        out0_lo_freq: float
        out0_lo_en: bool
        out0_att: int  # must be a multiple of 2!
        out0_offset_path0: float
        out0_offset_path1: float
        out1_lo_freq: float
        out1_lo_en: bool
        out1_att: int  # must be a multiple of 2!
        out1_offset_path0: float
        out1_offset_path1: float
        out_offsets: list[float] = field(
            init=False, default_factory=list  # QCM-RF module doesn't have an `out_offsets` parameter
        )

    settings: QbloxQCMRFSettings
    # TODO: We should separate instrument settings and instrument parameters, such that the user can quickly get
    # al the settable parameters of an instrument.
    parameters = {
        Parameter.OUT0_LO_FREQ,
        Parameter.OUT0_LO_EN,
        Parameter.OUT0_ATT,
        Parameter.OUT0_OFFSET_PATH0,
        Parameter.OUT0_OFFSET_PATH1,
        Parameter.OUT1_LO_FREQ,
        Parameter.OUT1_LO_EN,
        Parameter.OUT1_ATT,
        Parameter.OUT1_OFFSET_PATH0,
        Parameter.OUT1_OFFSET_PATH1,
    }

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """Initial setup"""
        super().initial_setup()
        for parameter in self.parameters:
            self.setup(parameter, getattr(self.settings, parameter.value))

    @Instrument.CheckDeviceInitialized
    def setup(
        self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None, port_id: str | None = None
    ):
        """Set a parameter of the Qblox QCM-RF module.

        Args:
            parameter (Parameter): Parameter name.
            value (float | str | bool): Value to set.
            channel_id (int | None, optional): ID of the sequencer. Defaults to None.
        """
        if parameter == Parameter.LO_FREQUENCY:
            if channel_id is not None:
                sequencer: AWGQbloxSequencer = self._get_sequencer_by_id(channel_id)
            elif port_id is not None:
                sequencer = self.get_sequencers_from_chip_port_id(chip_port_id=port_id)[0]
            else:
                raise ParameterNotFound(
                    "`channel_id` cannot be None when setting the `LO_FREQUENCY` parameter."
                    "Please specify the sequencer index or use the specific Qblox parameter."
                )

            # Remember that a set is ordered! Thus `{1, 0} == {0, 1}` returns True.
            # For this reason, the following checks also take into account swapped paths!
            if {sequencer.output_i, sequencer.output_q} == {0, 1}:
                output = 0
            elif {sequencer.output_i, sequencer.output_q} == {2, 3}:
                output = 1
            else:
                raise ValueError(
                    f"Cannot set the LO frequency of sequencer {channel_id} because it is connected to two LOs. "
                    f"The paths of the sequencer are mapped to outputs {sequencer.output_i} and {sequencer.output_q} "
                    "respectively."
                )
            parameter = Parameter(f"out{output}_lo_freq")

        if parameter in self.parameters:
            setattr(self.settings, parameter.value, value)
            self.device.set(parameter.value, value)
            return
        super().setup(parameter, value, channel_id)

    def get(self, parameter: Parameter, channel_id: int | None = None, port_id: str | None = None):
        """Set a parameter of the Qblox QCM-RF module.

        Args:
            parameter (Parameter): Parameter name.
            value (float | str | bool): Value to set.
            channel_id (int | None, optional): ID of the sequencer. Defaults to None.
        """
        if parameter == Parameter.LO_FREQUENCY:
            if channel_id is not None:
                sequencer: AWGQbloxSequencer = self._get_sequencer_by_id(channel_id)
            elif port_id is not None:
                sequencer = self.get_sequencers_from_chip_port_id(chip_port_id=port_id)[0]
            else:
                raise ParameterNotFound(
                    "`channel_id` cannot be None when setting the `LO_FREQUENCY` parameter."
                    "Please specify the sequencer index or use the specific Qblox parameter."
                )
            # Remember that a set is ordered! Thus `{1, 0} == {0, 1}` returns True.
            # For this reason, the following checks also take into account swapped paths!
            if {sequencer.output_i, sequencer.output_q} == {0, 1}:
                output = 0
            elif {sequencer.output_i, sequencer.output_q} == {2, 3}:
                output = 1
            else:
                raise ValueError(
                    f"Cannot set the LO frequency of sequencer {channel_id} because it is connected to two LOs. "
                    f"The paths of the sequencer are mapped to outputs {sequencer.output_i} and {sequencer.output_q} "
                    "respectively."
                )
            parameter = Parameter(f"out{output}_lo_freq")

        if parameter in self.parameters:
            return getattr(self.settings, parameter.value)
        return super().get(parameter, channel_id)

    def to_dict(self):
        """Return a dict representation of an `QCM-RF` instrument."""
        dictionary = super().to_dict()
        dictionary.pop("out_offsets")
        return dictionary
